/* see cpk_debug.h */
/* Last edited on 2009-02-09 18:17:54 by stolfi */ 

#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>

#include <i2.h>
#include <r2.h>
#include <pqueue.h>

#include <cpk_debug.h>
#include <cpk_io.h>
#include <cpk_graph.h>
#include <cpk_basic.h>

void cpk_print_vertices(FILE *wr, char *title, r2_vec_t *V, double_vec_t *W, int nMax)
  { int nV = V->ne;
    int u;
    fprintf(wr, "\n--------------------\n");
    fprintf(wr, "%s (count = %d):\n", title, nV);
    for (u = 0; u < nV; u++)
      { if ((nMax < 0) || (u < nMax) || (u >= nV - nMax))
          { fprintf(wr, "  " VTX_FMT, u); 
            r2_t *up = &(V->e[u]);
            r2_gen_print(wr, up, XY_FMT, " = (", ",", ")");
            if (W != NULL) 
              { fprintf(wr, " W = " WT_FMT, W->e[u]); }
            fprintf(wr, "\n");
          }
        else if (u == nMax)
          { fprintf(wr, "  ...\n"); }
      }
    fprintf(wr, "\n--------------------\n");
    fprintf(wr, "\n");
  }

void cpk_print_edges(FILE *wr, char *title, i2_vec_t *E, r2_vec_t *V, int nMax)
  { int nE = E->ne;
    int ie;
    fprintf(wr, "\n--------------------\n");
    fprintf(wr, "%s (count = %d):\n", title, E->ne);
    for (ie = 0; ie < nE; ie++)
      { i2_t *Ei = &(E->e[ie]);
        int u = ORG(*Ei);
        int v = DST(*Ei);
        if ((nMax < 0) || (ie < nMax) || (ie >= nE - nMax))
          { i2_gen_print(wr, Ei, VTX_FMT, "  (", ",", ")");
            if (V != NULL)
              { r2_t *up = &(V->e[u]);
                r2_t *vp = &(V->e[v]);
                fprintf(wr, "  at  ");
                r2_gen_print(wr, up, XY_FMT, "(", ",", ")");
                fprintf(wr, " -- ");
                r2_gen_print(wr, vp, XY_FMT, "(", ",", ")");
                fprintf(wr, "  d = %8.5f", r2_dist(up, vp));
              }
            fprintf(wr, "\n");
          }
        else if (ie == nMax)
          { fprintf(wr, "  ...\n"); }
      }
    fprintf(wr, "\n--------------------\n");
    fprintf(wr, "\n");
  }

void cpk_print_vertex_set(FILE *wr, char *title, int nJ, int *J, r2_vec_t *V, int nMax)
  { 
    int i;
    fprintf(wr, "\n--------------------\n");
    fprintf(wr, "%s (size = %d):%s", title, nJ, (V == NULL ? "" : "\n"));
    for(i = 0; i < nJ; i++) 
      { if ((nMax < 0) || (i < nMax) || (i >= nJ - nMax))
          { int u = J[i];
            if (V == NULL)
              { fprintf(wr, " " VTX_FFMT, u); }
            else
              { fprintf(wr, "  " VTX_FMT, u);
                r2_t *up = &(V->e[u]);
                fprintf(wr, "  at  ");
                r2_gen_print(wr, up, XY_FMT, "(", ",", ")");
                fprintf(wr, "\n");
              }
          }
        else if (i == nMax)
          { if (V == NULL)
              { fprintf(wr, " ... "); }
            else
              { fprintf(wr, "  ...\n"); }
          }
      }
    if (V == NULL) { fprintf(wr, "\n"); }
    fprintf(wr, "\n--------------------\n");
    fprintf(wr, "\n");
  }


void TRACE_SOL(char *title, int nS, double WS)
  { fprintf(stderr, "%s: tam = %d peso = " WT_FFMT "\n", title, nS, WS); }

void TRACE_Q(char *title, pqueue_t *Q)
  { fprintf(stderr, " %s:", title);
    int nQ = pqueue_count(Q); int i;
    for (i = 0; i < nQ; i++)
      { int u = pqueue_item(Q,i); 
        double eu = pqueue_value(Q,u); 
        fprintf(stderr, " %d:" WT_FFMT, u, eu); 
      }
    fprintf(stderr, "\n");
  }
