#ifndef cpk_debug_H
#define cpk_debug_H

/* Debugging routines for circle packing. */
/* Last edited on 2009-02-10 09:18:04 by stolfi */ 

#define _GNU_SOURCE
#include <stdio.h>

#include <i2.h>
#include <r2.h>
#include <pqueue.h>

#include <cpk_basic.h>

/* 
  DEBUGGING PRINTOUTS
  
  In all these procedures, if {nMax} is not negative, 
  prints only the first and last {nMax} vertices. */

void cpk_print_vertices(FILE *wr, char *title, r2_vec_t *V, double_vec_t *W, int nMax);
  /* Prints vertices {V.e[0..]} and (if {W} is not NULL) their
    weights {W[0..]} to file {wr}. */

void cpk_print_edges(FILE *wr, char *title, i2_vec_t *E, r2_vec_t *V, int nMax);
  /* Prints edges {E.e[0..]} to file {wr}, showing the indices and
    (if {V} is not null) the coordinates of the two endpoints. */

void cpk_print_vertex_set(FILE *wr, char *title, int nJ, int *J, r2_vec_t *V, int nMax);
  /* Prints the vertex list {J[0..nJ-1]} to {wr}.  Iv {V} is not null, prints also 
   their coordinates. */

/* OPTIMIZATION TRACING */

void TRACE_SOL(char *title, int nS, double WS);

void TRACE_Q(char *title, pqueue_t *Q);

/* PROGRAM TRACING */

#define cpk_trace(ARGS) \
  do { \
    fprintf(stderr, "\n[%s at %s:%d]", __FUNCTION__, __FILE__, __LINE__); \
    fprintf(stderr, " " ARGS);  fprintf(stderr, "\n"); \
  } while (0)

#define cpk_trace_entry(ARGS) \
  do { \
    fprintf(stderr, "\n[++ enter %s at %s:%d]", __FUNCTION__, __FILE__, __LINE__); \
    fprintf(stderr, " " ARGS);  fprintf(stderr, "\n"); \
  } while (0)

#define cpk_trace_exit(ARGS) \
  do { \
    fprintf(stderr, "\n[-- exit  %s at %s:%d]", __FUNCTION__, __FILE__, __LINE__); \
    fprintf(stderr, " " ARGS);  fprintf(stderr, "\n"); \
  } while (0)

#endif
