/* See cpk_graph.h */
/* Last edited on 2009-02-09 18:41:30 by stolfi */ 

#include <cpk_graph.h>

#include <bool.h>
#include <i2.h>

#include <stdlib.h>
#include <assert.h>
#include <affirm.h>

cpk_graph_t cpk_gather_neighbors(int nV, i2_vec_t E)
  {
    int nE = E.ne;

    /* Allocate areas: */
    cpk_graph_t G;
    G.nV = nV;
    G.nE = nE;
    G.nbr = (int *)notnull(malloc(2*nE*sizeof(int)), "no mem");
    G.deg = (int *)notnull(malloc(nV*sizeof(int)), "no mem");
    G.fnb = (int *)notnull(malloc((nV+1)*sizeof(int)), "no mem");
    
    /* Compute {G.deg[u]} for all {u}: */
    { int u; for (u = 0; u< nV; u++) { G.deg[u] = 0; } }
    { int j;
      for (j = 0; j < nE; j++) 
        { i2_t *Ej = &(E.e[j]);
          int u = ORG(*Ej), v = DST(*Ej); 
          /* Sanity check: */
          assert((u >= 0) && (u < nV));
          assert((v >= 0) && (v < nV)); 
          /* No self-loops: */
          assert(u != v);
          G.deg[u]++; G.deg[v]++;
        }
    }
    
    /* Compute {G.fnb[u] = SUM{G.deg[0..u-1]}, for {u=0..nV}: */
    { int k = 0;
      int u; 
      for (u = 0; u < nV; u++) { G.fnb[u] = k; k += G.deg[u]; }
      G.fnb[nV]=k;
      assert(k==2*nE);
    }
 
    /* Gather neighbors of each vertex: */
    { int u; for (u = 0; u< nV; u++) { G.deg[u] = 0; } }
    { int j; 
      for (j = 0; j < nE; j++) 
        { i2_t *Ej = &(E.e[j]);
          int u = ORG(*Ej), v = DST(*Ej); 
          assert(u != v);
          G.nbr[G.fnb[u]+G.deg[u]] = v; G.deg[u]++; 
          G.nbr[G.fnb[v]+G.deg[v]] = u; G.deg[v]++;
        }
    }
    return G;
  }

bool_t cpk_graph_adjacent(cpk_graph_t *G, int u, int v)
  { if (u == v) { return FALSE; /* Graphs are assumed loopfree. */ } 
    int du = G->deg[u];
    int dv = G->deg[v];
    int i;
    if (du < dv)
      { /* Look for {v} among the neighbors of {u}: */
        int *nbu = &(G->nbr[G->fnb[u]]); 
        for (i = 0; i < du; i++) { if (nbu[i] == v) { return TRUE; } } 
      }
    else
      { /* Look for {u} among the neighbors of {v}: */
        int *nbv = &(G->nbr[G->fnb[v]]);
        for (i = 0; i < dv; i++) { if (nbv[i] == u) { return TRUE; } } 
      }
    return FALSE;
  }

int cpk_graph_count_common_neighbors(cpk_graph_t *G, int u, int v, bool_t temp[])
  { int du = G->deg[u];
    int dv = G->deg[v];
    if (u == v) { return du; } 
    int i;
    int count = 0;
    /* Mark all neighbors of {u}: */
    int *nbu = &(G->nbr[G->fnb[u]]); 
    for (i = 0; i < du; i++) { int w = nbu[i]; assert(!temp[w]); temp[w] = TRUE; } 
    /* Count neighbors of {v} that are marked: */
    int *nbv = &(G->nbr[G->fnb[v]]);
    for (i = 0; i < dv; i++) { if (temp[nbv[i]]) { count++; } } 
    /* Be a good boy and clean up after yourself: */
    for (i = 0; i < du; i++) { temp[nbu[i]] = FALSE; } 
    return count;
  }

