#ifndef cpk_grasp_H
#define cpk_grasp_H

/* Definies gerais para soluo de MIS com GRASP. */
/* Last edited on 2007-01-03 23:26:25 by stolfi */ 

#include <cpk_basic.h>
#include <cpk_graph.h>

#include <bool.h>
#include <r2.h>
#include <vec.h>

/* 4. Constantes usadas no GRASP */
#define cpk_grasp_REL_WT_NOISE 0.10 
  /* Perturbao relativa mxima no peso para randomizao da construo. */

int_vec_t cpk_grasp_find_indep_set
  ( cpk_graph_t *G,   /* The incompatibilty graph. */
    double *W,        /* Weight of each vertex. */
    double maxClock,  /* Deadline for result. */
    int seed,         /* Seed for random choices. */
    bool_t verbose    /* TRUE prints debugging diagnostics. */
  );
  /* Computa um conjunto independente (IS) de vrtices de {G} com peso
    mximo (espera-se), usando uma adaptao do algoritmo GRASP
    proposto por Resende (``A Greedy Adaptative Search Procedure for
    the Maximum Independent Set'', Operations Research, ...).
    
    O procedimento repetidamente constri uma soluo maximal {S},
    (/fase de construo/), e aplica a cada uma delas um procedimento
    de otimizao local (/fase de busca/). Este processo termina
    quando o relgio de CPU {now()} passar de {maxClock}.
    
    A fase de construo inicializa {S} com dois vrtices "bons" e
    acrescenta mais vrtices, um a um, at {S} ficar maximal. Os
    vrtices so escolhidos de forma gulosa (pelo critrio da razo
    benefcio/custo), mas com uma certa perturbao aleatria.
    
    A fase de busca repetidamente retira at {maxRem} vrtices da
    soluo {S}, e tenta cobrir o "buraco" assim criado, de maneira
    gulosa ou exaustiva. */

#endif
