/* see cpk_io.h */
/* Last edited on 2009-02-10 08:05:29 by stolfi */ 

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <bool.h>
#include <interval.h>
#include <i2.h>
#include <r2.h>
#include <pswr.h>
#include <fget.h>
#include <affirm.h>
#include <jsfile.h>

#include <cpk_io.h>
#include <cpk_main.h>
#include <cpk_coords.h>
#include <hxg_ps.h>
#include <cpk_basic.h>

void cpk_plot_solution
  ( char *outDir,
    char *solTag,
    char *demTag,
    interval_t B[],
    cpk_domain_t *C,
    cpk_policy_t *P,
    r2_vec_t *V,
    double_vec_t *W,
    int_vec_t *J
  )
  {
    /* Compose output file name: */
    char *fileName = NULL;
    asprintf(&fileName, "%s/%s-%s", outDir, solTag, demTag);
    
    /* Plot the domain and the candidates: */
    fprintf(stderr, "Writing Postscript file \"%s.ps\" ...\n", fileName);
    PSStream *ps = hxg_ps_init_plot(B, fileName, FALSE, "letter");
    
    /* Show the domain: */
    cpk_plot_domain(ps, C, P);
    
    /* Show the demand points and their adequacy radii: */
    cpk_plot_demand(ps, &(C->Dem), P->rAuc);

    /* Show candidate points: */
    cpk_plot_candidates(ps, V, W);
    
    /* Show proposed auctions: */
    cpk_plot_proposed_auctions(ps, J, V, P->rAuc, P->dMin/2);
    
    /* Show statistics and such: */
    cpk_plot_solution_attributes(ps, B, P, W, J);

    /* A frame: */
    pswr_set_pen(ps, 0.8,0.5,0.2, 0.10, 0,0);
    pswr_frame(ps);
    
    pswr_close_stream(ps);
  }
  
r2_vec_t cpk_read_points(char *inDir, char *tag)
  {
    FILE *rd;
    { char *path = NULL;
      asprintf(&path, "%s/%s.txt", inDir, tag);
      rd = open_read(path, TRUE);
      free(path);
    }
    
    /* Read file. */
    r2_vec_t P = r2_vec_new(100);
    int nP = 0;
    while(1)
      { /* Skip blanks on same line:: */
        fget_skip_spaces(rd);
        /* Check next character: */
        int c = fgetc(rd);
        if (c == EOF)
          { break; }
        else if (c == '\n')
          { /* Blank line, ignore: */ }
        else if (c == '\015')
          { /* CR: ignore, like LF, but eat next LF if any: */
            c = fgetc(rd);
            if (c == EOF) { break; }
            if (c != '\n') { ungetc(c, rd); }
          }
        else if 
          ( (c == '+') || (c == '-') || (c == '.') || 
            ((c >= '0') && (c <= '9'))
          )
          { /* Number, must be a point: */
            ungetc(c, rd);
            double Lon = fget_double(rd);
            double Lat = fget_double(rd);
            r2_t p = (r2_t){{Lon,Lat}};
            /* Rest of line must be empty: */
            fget_eol(rd);
            /* Save point in list: */
            r2_vec_expand(&P, nP);
            P.e[nP] = p; 
            nP++;
          }
        else 
          { /* Non-number, non-blank: assume it is a comment, ignore: */
            do { c = fgetc(rd); } while ((c != '\n') && (c != EOF));
            if (c == EOF) { break; }
          }
      }
    /* Trim: */
    r2_vec_trim(&P, nP);
    return P;
  }

r2_vec_t cpk_read_polys(char *inDir, char *tag, bool_t closed)
  {
    FILE *rd;
    { char *path = NULL;
      asprintf(&path, "%s/%s.txt", inDir, tag);
      rd = open_read(path, TRUE);
      free(path);
    }
    
    /* Read file. */
    r2_vec_t P = r2_vec_new(100);
    int nP = 0;
    int fP = -1; /* First point of current component, or -1 outside components. */
    while(1)
      { /* Skip blanks on same line:: */
        fget_skip_spaces(rd);
        /* Check next character: */
        int c = fgetc(rd);
        if (c == EOF)
          { break; }
        else if (c == '\n')
          { /* Blank line, ignore: */ }
        else if (c == '\015')
          { /* CR: ignore, like LF, but eat next LF if any: */
            c = fgetc(rd);
            if (c == EOF) { break; }
            if (c != '\n') { ungetc(c, rd); }
          }
        else if 
          ( (c == '+') || (c == '-') || (c == '.') || 
            ((c >= '0') && (c <= '9'))
          )
          { /* Number, must be a point: */
            ungetc(c, rd);
            double Lon = fget_double(rd);
            double Lat = fget_double(rd);
            r2_t p = (r2_t){{Lon,Lat}};
            /* Rest of line must be empty: */
            fget_eol(rd);
            /* Check component status: */
            if (fP == -1)
              { /* First point of a component. */
                if (nP > 0)
                  { /* Not the first component, must insert {{#}-vertex: */
                    r2_vec_expand(&P, nP);
                    P.e[nP] = (r2_t){{INF,INF}}; 
                    nP++;
                  }
                /* Remember this point: */
                fP = nP;
              }
            else if (closed)
              { /* Check for component closure: */
                if (r2_eq(&p, &(P.e[fP])))
                  { /* End of component: */
                    fP = -1; 
                  }
              }
            /* Save point in list: */
            r2_vec_expand(&P, nP);
            P.e[nP] = p; 
            nP++;
          }
        else 
          { /* Non-number, non-blank: assume it is a component separator: */
            if ((fP != -1) && closed) 
              { demand(FALSE, "component is not closed"); }
            fP = -1;
            do { c = fgetc(rd); } while ((c != '\n') && (c != EOF));
            if (c == EOF) { break; }
          }
      }
    /* End of input, check for proper closure: */
    if ((fP != -1) && closed)
      { demand(FALSE, "component is not closed"); }
    /* Trim: */
    r2_vec_trim(&P, nP);
    return P;
  }
  
void cpk_pick_ref_coords(r2_vec_t *P, double *refLon, double *refY)
  { 
    /* Simple-minded: take the averages over all finite vertices.
      Should be OK for any region in Brazil. For international 
      use we should check for wrap-around in longitude
      and/or in Northing. Some day ...
    */
    /* Compute the average longitude: */
    double sLon = 0;
    int nP = 0;
    int i;
    for (i = 0; i < P->ne; i++) 
      { r2_t *pi = &(P->e[i]); 
        if (r2_is_finite(pi))
          { sLon += X(*pi); nP++; } 
      }
    /* If there are no finite points, return (0,0): */
    if (nP <= 0) { (*refLon) = 0; (*refY) = 0; return; }
    /* The reference longitude is the mean vertex longitude. */
    (*refLon) = sLon/nP;
    /* Map all points to EUTM, and compute the average Y: */
    double sY = 0; 
    for (i = 0; i < P->ne; i++) 
      { r2_t *Pi = &(P->e[i]); 
        if (r2_is_finite(Pi))
          { double Lon = X(*Pi);
            double Lat = Y(*Pi);
            double x, y;
            LL_to_EUTM(Lat, Lon, &x, &y, *refLon);
            sY += y;
          } 
      }
    (*refY) = sY/nP;
  }

r2_vec_t cpk_LL_to_EUTM(r2_vec_t *P, double refLon, double refY, double magnify)
  { r2_vec_t Q = r2_vec_new(P->ne);
    int i;
    for (i = 0; i < P->ne; i++) 
      { r2_t *Pi = &(P->e[i]); 
        r2_t *Qi = &(Q.e[i]); 
        if (r2_is_finite(Pi))
          { double Lon = X(*Pi);
            double Lat = Y(*Pi);
            double x, y;
            LL_to_EUTM(Lat, Lon, &x, &y, refLon);
            (*Qi) = (r2_t){{x*magnify, (y - refY)*magnify}};
          }  
        else
          { (*Qi) = (*Pi); }
      }
    return Q;
  }

r2_vec_t cpk_EUTM_to_LL(r2_vec_t *P, double refLon, double refY, double magnify)
  { r2_vec_t Q = r2_vec_new(P->ne);
    int i;
    for (i = 0; i < P->ne; i++) 
      { r2_t *Pi = &(P->e[i]); 
        r2_t *Qi = &(Q.e[i]); 
        if (r2_is_finite(Pi))
          { double x = X(*Pi)/magnify;
            double y = Y(*Pi)/magnify;
            double Lat, Lon;
            EUTM_to_LL(x, y + refY, &Lat, &Lon, refLon);
            (*Qi) = (r2_t){{Lon,Lat}};
          } 
        else
          { (*Qi) = (*Pi); }
      }
    return Q;
  }

void cpk_plot_domain(PSStream *ps, cpk_domain_t *C, cpk_policy_t *P)
  {
    /* Paint the polygon: */
    pswr_set_fill_color(ps, 1.00,0.95,0.93);
    hxg_ps_fill_polygon(ps, &(C->Urb));

    /* Unpaint the forbidden zones around existing/planned stations: */
    pswr_set_fill_color(ps, 1.00,1.00,1.00);
    hxg_ps_fill_circles(ps, &(C->Exs), NULL, P->dMin + P->rAuc);
    
    /* Unpaint the forbidden zones around predetermined auction centers: */
    pswr_set_fill_color(ps, 1.00,1.00,1.00);
    hxg_ps_fill_circles(ps, &(C->Auc), NULL, P->rAuc + P->dMin + P->rAuc);
    
    /* Unpaint the forbidden zone near intermunicipal borders: */
    hxg_ps_fill_sausage(ps, &(C->Mun), P->dMun);

    /* Unpaint the forbidden zone near international borders: */
    hxg_ps_fill_sausage(ps, &(C->Nat), P->dNat);

    /* Draw the polygon outline: */
    pswr_set_pen(ps, 0.50,0.25,0.00, 0.10, 0,0);
    hxg_ps_draw_polyline(ps, &(C->Urb));
    
    /* Draw the intermunicipal borders: */
    pswr_set_pen(ps, 0.50,0.00,1.00, 0.20, 0,0);
    hxg_ps_draw_polyline(ps, &(C->Mun));

    /* Draw the international borders: */
    pswr_set_pen(ps, 1.00,0.00,0.50, 0.30, 0,0);
    hxg_ps_draw_polyline(ps, &(C->Nat));

    /* Draw the exs./pln. stations: */
    hxg_color_t colorS = (hxg_color_t){ 0.00, 0.30, 1.00 };
    cpk_plot_stations(ps, &(C->Exs), 0.0, P->dMin/2, colorS);

    /* Draw the predetermined auction points: */
    hxg_color_t colorA = (hxg_color_t){ 0.50, 0.30, 0.00 };
    cpk_plot_stations(ps, &(C->Auc), P->rAuc, P->dMin/2, colorA);
  }

void cpk_plot_stations
  ( PSStream *ps, 
    r2_vec_t *P, 
    double rUnc, 
    double rRad,
    hxg_color_t color
  )
  { 
    hxg_color_t z = color;
    
    /* Uncertainty circles (solid): */
    pswr_set_pen(ps, z.R,z.G,z.B, 0.10, 0,0);
    hxg_ps_draw_circles(ps, P, NULL, rUnc); 
    
    /* Nominal broadcast coverage areas (dashed): */
    pswr_set_pen(ps, z.R,z.G,z.B, 0.10, 1.0,0.5);
    hxg_ps_draw_circles(ps, P, NULL, rUnc+rRad); /* . */
    
    /* Station positions or auction centers (dots): */
    pswr_set_fill_color(ps, z.R,z.G,z.B);
    hxg_ps_fill_dots(ps, P, 0.3);
    
    if (P->ne < 1000)
      { /* Write station indices: */
        r2_t disp = (r2_t){{ rRad/8, 0 }};
        pswr_set_pen(ps, z.R,z.G,z.B, 0.10, 0,0);
        hxg_ps_show_labels(ps, P, disp, 0.0,0.5, 8);
      }
  }

void cpk_plot_demand(PSStream *ps, r2_vec_t *P, double rDem)
  { 
    hxg_color_t z = (hxg_color_t){ 0.00, 0.60, 0.00 };
    
    /* Circle of adequate auction centers: */
    /* pswr_set_pen(ps, z.R,z.G,z.B, 0.10, 0,0); */
    /* hxg_ps_draw_circles(ps, P, NULL, rDem);  */
    
    /* Points of demand: */
    pswr_set_fill_color(ps, z.R,z.G,z.B);
    hxg_ps_fill_dots(ps, P, 0.5);
  }
  
void cpk_plot_candidates(PSStream *ps, r2_vec_t *V, double_vec_t *W)
  {
    hxg_color_t z = (hxg_color_t){ 0.7, 0.4, 0.0 };
    
    int nV = V->ne;
    int i;
    pswr_set_pen(ps, 0.0,0.0,0.0, 0.10, 0.0, 0.0);
    for (i = 0; i < nV; i++)
      { r2_t *Vi = &(V->e[i]);
        double Ni = (i < W->ne ? W->e[i] : 0);
        double r = 0.10 + 0.05*floor(log(Ni+1)/log(2));
        if (Ni == 0)
          { pswr_set_fill_color(ps, 0.0,0.0,0.0); }
        else
          { pswr_set_fill_color(ps, z.R,z.G,z.B); }
        pswr_dot(ps, X(*Vi), Y(*Vi), r, TRUE, FALSE);
      }
  }

void cpk_plot_proposed_auctions
  ( PSStream *ps, 
    int_vec_t *J, 
    r2_vec_t *V, 
    double rAuc, 
    double rRad
  )
  {
    int nV = V->ne;
    int nJ = J->ne;
    
    /* Extract the auction centers {U} with uncertainties {hU}: */
    int nU = nJ;
    r2_vec_t U = r2_vec_new(nU);
    { int j;
      for (j = 0; j < nJ; j++)
        { int i = J->e[j]; 
          assert(i >= 0);
          assert(i < nV);
          U.e[j] = V->e[i];
        }
    }
    
    /* Now plot them: */
    hxg_color_t colorU = (hxg_color_t){ 1.0, 0.3, 0.0 };
    cpk_plot_stations(ps, &U, rAuc, rRad, colorU);
    free(U.e);
  }

void cpk_plot_solution_attributes
  ( PSStream *ps, 
    interval_t B[], 
    cpk_policy_t *P,
    double_vec_t *W,
    int_vec_t *J
  )
  {
    int nJ = J->ne;
    
    /* Set font and choose position for key: */
    pswr_set_pen(ps, 0.0,0.0,0.5, 0.10, 0.0, 0.0);
    double ptSize = 10;
    pswr_set_label_font(ps, "Courier", ptSize);
    
    /* Print the solution size: */
    { char *msg = NULL;
      asprintf(&msg, "Num auctions = %d", nJ);
      pswr_add_caption(ps, msg, 0.0);
      free(msg);
    }
    
    /* Print the solution weight: */
    { double WJ = 0.0;
      int i; 
      for (i = 0; i < nJ; i++) { WJ += W->e[J->e[i]]; }
      char *msg = NULL;
      asprintf(&msg, "Total weight = " WT_FFMT, WJ);
      pswr_add_caption(ps, msg, 0.0);
      free(msg);
    }
  }

