/* See cpk_lopt.h */
/* Last edited on 2007-01-03 23:30:24 by stolfi */ 

#include <cpk_lopt.h>
#include <cpk_basic.h>
#include <cpk_lopt_simple.h>
#include <cpk_lopt_multi.h>
#include <cpk_graph.h>
#include <cpk_mis.h>
#include <cpk_valid.h>
#include <cpk_debug.h>
#include <cpk_io.h>

#include <bool.h>
#include <vec.h>

#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <assert.h>

/* PROTOTIPOS INTERNOS */

/* IMPLEMENTAES */

void cpk_local_opt
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int maxRem,        /* Nmero mximo de vrtice a trocar em cada tentativa. */
    int *nSM,          /* Tamanho da melhor soluo j vista. */
    int SM[],          /* {SM[0..nSM-1]}  a melhor soluo j vista. */
    double *WSM,       /* Peso total da melhor soluo {SM}. */
    bool_t verbose     /* TRUE para imprimir mensagens de diagnstico. */
  )
  { if (maxRem == 1) 
      { cpk_local_opt_simple(SQ, nSM, SM, WSM, verbose); }
    else
      { cpk_local_opt_multiple(SQ, maxRem, nSM, SM, WSM, verbose); }
  }

void cpk_lopt_update_best_solution
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int *nSM,          /* Tamanho da melhor soluo j vista. */
    int SM[],          /* {SM[0..nSM-1]}  a melhor soluo j vista. */
    double *WSM,       /* Peso total da melhor soluo {SM}. */
    bool_t verbose     /* TRUE para imprimir mensagens de diagnstico. */
  )
  {
    if (cpk_weight_cmp(SQ->WS, *WSM) > 0)
      { /* O resultado  melhor que {SM}; copia {S} para {SM}: */
        { int i;  for(i = 0; i < SQ->nS; i++) { SM[i] = SQ->S[i]; } }
        (*nSM) = SQ->nS;
        (*WSM) = SQ->WS; 
        if (verbose)
          { TRACE_SOL("  Melhorou soluo global", (*nSM), (*WSM));
            cpk_print_vertex_set(stderr, "", (*nSM), SM, NULL, 30);
            if (verbose) { fprintf(stderr, "\n");  }
          }
      }
  }

void cpk_lopt_restore_solution
  ( cpk_mis_state_t *SQ,     /* Soluo corrente {S} e estruturas auxiliares. */
    int nDel,        /* Nmero de vrtices a retirar. */
    int nAdd,        /* Nmero de vrtices a recolocar. */
    int Add[],       /* {Add[0..nAdd-1]} so os vrtices a recolocar. */
    bool_t verbose   /* TRUE para imprimir mensagens de diagnstico. */
  )
  {
    if (nDel > 0)
      { if (verbose) { fprintf(stderr, "      Retirando os ltimos %d vrtices", nDel); }
        int nSred = SQ->nS - nDel;  /* Nmero de vrtices a deixar: */
        while(SQ->nS > nSred) { int v = SQ->S[SQ->nS-1];  cpk_mis_delete_indep(SQ, v, verbose); }
        if (verbose) { fprintf(stderr, "\n");  }
      }
    if (nAdd > 0)
      { if (verbose) { fprintf(stderr, "      Recolocando %d vrtices", nAdd); }
        while (nAdd > 0) { nAdd--; int v = Add[nAdd]; cpk_mis_add_indep(SQ, v, verbose); }
        if (verbose) { fprintf(stderr, "\n");  }
      }
  }

bool_t cpk_lopt_same_set(int nX, int X[], int nY, int Y[])
  {
    if (nX != nY) { return FALSE; }
    int iX, iY;
    for (iX = 0; iX < nX; iX++)
      { for (iY = 0; (iY < nY) && (X[iX] != Y[iY]); iY++) { }
        if (iY >= nY) { return FALSE; }
      }
    return TRUE;
  }
