#ifndef cpk_lopt_H
#define cpk_lopt_H

/* Otimizao local gulosa de um conjunto independente maximal. */
/* Last edited on 2007-01-03 23:27:04 by stolfi */ 

#include <cpk_graph.h>
#include <cpk_mis.h>
#include <cpk_basic.h>

#include <bool.h>

void cpk_local_opt
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int maxRem,        /* Nmero mximo de vrtice a trocar em cada tentativa. */
    int *nSM,          /* Tamanho da melhor soluo j vista. */
    int SM[],          /* {SM[0..nSM-1]}  a melhor soluo j vista. */
    double *WSM,       /* Peso total da melhor soluo {SM}. */
    bool_t verbose     /* TRUE para imprimir mensagens de diagnstico. */
  );
  /* Supe que o conjunto {S} armazenado em {SQ}  um conjunto
    independente maximal. Examina alguns outros conjuntos maximais
    {S'} derivados de {S} pela retirada de um certo nmero {nRem}
    de vrtices e re-expanso gulosa para conjunto maximal. 
    
    As variveis {SM[0..nSM-1]} e {WSM} devem conter respectivamente
    os vrtices e o peso total da melhor soluo j examinada at o
    momento. Elas so atualizadas sempre que o procedimento encontra
    uma soluo {S} melhor que {SM}.
    
    O procedimento faz vrias passadas sobre o grafo, comeando com
    {nRem=1}, depois com {nRem=2}, etc; e voltando para {nRem=1} cada
    vez que consegue uma melhora. O procedimento retorna quando uma
    passada sobre todos os vrtices e todos os {nRem} de 1 a {maxRem}
    no consegue melhorar a soluo. */

bool_t cpk_lopt_same_set(int nX, int X[], int nY, int Y[]);
  /* Verdadeiro se e somente se o conjunto {X[0..nX-1]}  igual a {Y[0..nY-1]},
    ignorando a ordem. til para verificar se vrtices recolocados
    na soluo so os que foram tirados. Tempo: O(nX*nY). */

void cpk_lopt_update_best_solution
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int *nSM,          /* Tamanho da melhor soluo j vista. */
    int SM[],          /* {SM[0..nSM-1]}  a melhor soluo j vista. */
    double *WSM,       /* Peso total da melhor soluo {SM}. */
    bool_t verbose     /* TRUE para imprimir mensagens de diagnstico. */
  );
  /* Se o peso {WS} da soluo corrente {S} guardada em {SQ}  maior
    que {WSM}, copia {nS,S} para {nSM,SM}, e atualiza {WSM}. */
    
void cpk_lopt_restore_solution
  ( cpk_mis_state_t *SQ,     /* Soluo corrente {S} e estruturas auxiliares. */
    int nDel,        /* Nmero de vrtices a retirar. */
    int nAdd,        /* Nmero de vrtices a recolocar. */
    int Add[],       /* {Add[0..nAdd-1]} so os vrtices a recolocar. */
    bool_t verbose   /* TRUE para imprimir mensagens de diagnstico. */
  );
  /* Modifica a soluo {S} retirando dela os ltimos {nDel} vrtices,
    e acrescentando os vrtices {Add[0..nAdd-1]}, que ficaro no final
    de {S}. Atualiza o peso total {WS} e a fila {Q} adequadamente. */

#endif
