/* See cpk_lopt_multi.h */
/* Last edited on 2008-01-09 00:32:27 by stolfi */ 

#include <cpk_lopt_multi.h>
#include <cpk_basic.h>
#include <cpk_lopt.h>
#include <cpk_greedy.h>
#include <cpk_graph.h>
#include <cpk_mis.h>
#include <cpk_weight.h>
#include <cpk_valid.h>
#include <cpk_debug.h>
#include <cpk_io.h>

#include <bool.h>
#include <vec.h>

#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <assert.h>

/* PROTOTIPOS INTERNOS */

bool_t cpk_local_opt_at_pivot_1
  ( cpk_mis_state_t *SQ,     /* Soluo corrente {S} e estruturas auxiliares. */
    int nRem,        /* Nmero de vrtices a trocar em cada tentativa. */
    int x,           /* Vrtice piv. */
    bool_t verbose,  /* TRUE para imprimir mensagens de diagnstico. */
    bool_t piv[],    /* {piv[v] = TRUE} se vale a pena tentar {v} como piv. */
    int T[],
    bool_t temp[]
  );
  /* Se {x} est em {S} ou no tem exatamente {nRem} vizinhos em {S},
    no faz nada e retorna FALSE. Seno, tenta melhorar {S} retirando todos
    os vrtices {T} de {S} que so adjacentes a {x} (que no deve esta
    em {S}) e recobrindo o buraco assim aberto com um algoritmo guloso
    com lookahead. Se isso melhorar a soluo, devolve TRUE; seno
    restaura {S} como antes e devolve FALSE.
    
    O procedimento tambm faz {piv[v]= FALSE} para {x} e todo vrtice
    {v} que ficou descoberto com a retirada de {T}, ou seja, para {x}
    e para todo {v} em {G-S} cujos vizinhos em {S} esto todos em {T}.
    
    Os parmetros {T} e {temp} so reas de trabalho que devem ser
    alocadas pelo cliente, com pelo menos {nV} elementos. O vetor
    {temp} deve ser todo FALSE na entrada, e ser todo FALSE na
    sada. */

bool_t cpk_local_opt_at_pivot_2
  ( cpk_mis_state_t *SQ,     /* Soluo corrente {S} e estruturas auxiliares. */
    int nRem,        /* Nmero de vrtices a trocar em cada tentativa. */
    int x,           /* Vrtice piv. */
    bool_t verbose,  /* TRUE para imprimir mensagens de diagnstico. */
    bool_t piv[],    /* {piv[v] = TRUE} se vale a pena tentar {v} como piv. */
    int T[],
    bool_t temp[]
  );
  /* Semelhante a {}, exceto que o piv {x}  includo na nova soluo como vrtice
    inicial da busca gulosa, e ele  o nico vrtice que  marcado com {piv = FALSE}. */

/* IMPLEMENTAES */

#define VALIDATE TRUE

void cpk_local_opt_multiple
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int maxRem,        /* Nmero mximo de vrtice a trocar em cada tentativa. */
    int *nSM,          /* Tamanho da melhor soluo j vista. */
    int SM[],          /* {SM[0..nSM-1]}  a melhor soluo j vista. */
    double *WSM,       /* Peso total da melhor soluo {SM}. */
    bool_t verbose     /* TRUE para imprimir mensagens de diagnstico. */
  )
  { if (verbose) { cpk_trace_entry(); }
    
    /* A estratgia  retirar de {S} grupos de {nRem} vrtices
    prximos, com {nRem} em {1..maxRem}, e depois completar {S}
    para maximal, gulosamente.

    A cada retirada, escolhe-se um vrtice /piv/ {x} de {G-S} com
    exatamente {nRem} vizinhos em {S}. Seja {T} o conjunto desses
    vizinhos. Retira-se {T} de {S}. Seja {Q} o conjunto dos vrtices
    de {G} que, depois dessa retirada, no esto em {S} nem so
    adjacentes a {S}; este conjunto no  vazio pois contm pelo
    menos {x}. O conjunto {S}  ento re-expandido gulosamente s
    custas de {Q}, at se tornar maximal.

    Para expandir {S-T}, usamos uma busca gulosa com lookahead (vide
    abaixo). Note que  possvel que o novo conjunto maximal seja pior
    que o anterior. Nesse caso a mexida  desfeita, restaurando-se o
    conjunto {S} original.

    Sempre que um vrtice {w}  includo em {Q}, ele  marcado
    fazendo-se {piv[w] = FALSE}, e excludo de
    futuras escolhas de piv para este {nRem}. Quando todos os 
    vrtices foram excludos, recomea-se tudo de novo com 
    {nRem = nRem+1} e {piv[0..nV-1] = TRUE}.
    Quando a soluo corrente  melhorada,
    recomea-se tudo de novo com {nRem = 1} e {piv[0..nV-1] = TRUE}. */

    cpk_graph_t *G = SQ->G;
    int nV = G->nV;
    
    /* Areas de trabalho: */
    bool_t piv[nV];
    int T[nV];
    bool_t temp[nV];
    { int v; for (v = 0; v < nV; v++) { temp[v] = FALSE; } }
    
    /* Inicializa o vetor de vrtices pivotveis: */
    { int v; for (v = 0; v < nV; v++) { piv[v] = TRUE; } }

    /* Repete busca na vizinhana enquanto houver melhoria. */
    int x = 0; /* ltimo vrtice de {G} usado como piv. */
    int nRem = 1; /* Nmero de {S}-vizinhos do piv a remover: */
    int nPass = 0; /* Nmero de tentativas que deram certo. */
    int nTentou = 0;  /* Nmero de tentativas fracassadas consecutivas p/ {nRem}. */
    do 
      { if (nTentou == 0) 
          { if (verbose) { fprintf(stderr, "\n  Passada %d nRem = %d\n\n", nPass, nRem); } }
        /* Pega prximo piv: */
        x--; if (x < 0) { x = nV-1; }

        /* Lembra do peso atual: */
        double WSold = SQ->WS;
        int nSold = SQ->nS;
        /* Tenta troca com {nRem} vrtices na vizinhana de {x}: */
        if (piv[x] && cpk_local_opt_at_pivot_1(SQ, nRem, x, verbose, piv, T, temp))
          { /* Mudou; deve ser para melhor: */
            assert(cpk_weight_cmp(WSold, SQ->WS) < 0);
            /* Tente atualizar a melhor soluo. */
            cpk_lopt_update_best_solution(SQ, nSM, SM, WSM, verbose);
            /* Comece tudo de novo: */ 
            nTentou = 0; nRem = 1; nPass++;
            /* Reinicializa o vetor {piv}. */
            { int v; for (v = 0; v < nV; v++) { piv[v] = TRUE; } }
          }
        else
          { /* No mudou; deve ter mantido peso e tamanho: */
            assert(nSold == SQ->nS);
            assert(cpk_weight_cmp(WSold, SQ->WS) == 0);
            /* Por via das dvidas: */
            piv[x] = FALSE;
            /* Mais uma tentativa fracassada: */
            nTentou++;
            /* Esgotamos este {nRem}? */
            if (nTentou >= nV)
              { /* Vamos para o seguinte: */
                nRem++; nTentou = 0; 
                /* Reinicializa {piv}: */
                { int v; for (v = 0; v < nV; v++) { piv[v] = TRUE; } }
              }
          } 
      } 
    while (nRem <= maxRem);
    if (VALIDATE) { cpk_check_indep_set(G, SQ->nS, SQ->S, NULL); }
    if (verbose) { cpk_trace_exit(); }
  }

bool_t cpk_local_opt_at_pivot_1
  ( cpk_mis_state_t *SQ,     /* Soluo corrente {S} e estruturas auxiliares. */
    int nRem,        /* Nmero de vrtices a trocar em cada tentativa. */
    int x,           /* Vrtice piv. */
    bool_t verbose,  /* TRUE para imprimir mensagens de diagnstico. */
    bool_t piv[],    /* {piv[v] = TRUE} se vale a pena tentar {v} como piv. */
    int T[],
    bool_t temp[]
  )
  {
    cpk_graph_t *G = SQ->G;
    pqueue_t *Q = SQ->Q;
    
    assert(piv[x]); /* S deve chamar se {x} ainda est marcado como possvel piv. */
    assert(pqueue_count(Q) == 0); /* Neste ponto {S} deveria ser maximal. */

    /* O vrtice {x} realmente satisfaz as condies para piv? */
    if ((SQ->locS[x] >= 0) || (SQ->degS[x] != nRem)) { return FALSE; } 
    
    if (verbose) { fprintf(stderr, "    Mexendo nas redondezas do vrtice %d\n", x); }

    /* Lembra do peso da soluo corrente: */
    double WSold = SQ->WS;

    /* Retire de {S} todos os vizinhos de {x}, salvando-os em {T[0..nT-1]}. */
    /* Coloque os vrtices que ficaram admissveis em {Q}. */
    if (verbose) { fprintf(stderr, "      Retirando %d vrtices:", SQ->degS[x]); }
    int nT = 0;
    { int dx = G->deg[x]; int *nbx = &(G->nbr[G->fnb[x]]);
      int i;
      for (i = 0; i < dx; i++)
        { int v = nbx[i]; 
          if (SQ->locS[v] >= 0) 
            { cpk_mis_delete_indep(SQ, v, verbose); T[nT] = v; nT++; }
        }
    }
    assert(SQ->degS[x] == 0); /* Objetivo da retirada. */
    if (verbose) { fprintf(stderr, "\n"); }

    /* Marca todos os vrtices de {Q} como no mais pivotveis: */
    int nQ = pqueue_count(Q);
    if (verbose) { fprintf(stderr, "      Criados %d vrtices admissveis:", nQ);}
    assert(nQ > 0); /* Pois pelo menos {x} deve ter ficado admissvel. */
    { int i;
      for (i = 0; i < nQ; i++)
        { int q = pqueue_item(Q, i);
          piv[q] = FALSE; 
          if (verbose) { fprintf(stderr, " %d", q);}
        }
    }
    if (verbose) { fprintf(stderr, "\n");}
    /* TRACE_Q("     Q =", Q); */

    /* Lembra posio corrente em {S}: */
    int nSred = SQ->nS; 

    /* Roda algoritmo greedy sobre {G[Q]}: */
    if (verbose) { fprintf(stderr, "      Expanso gulosa..."); }
    cpk_greedy_maximize_lookahead(SQ, verbose, temp);
    if (verbose) { fprintf(stderr, " recolocou %d vrtices\n", SQ->nS - nSred);  }

    /* Houve melhora? */
    if (cpk_weight_cmp(SQ->WS, WSold) > 0)
      { /* Houve melhora! */
        if (verbose) { TRACE_SOL("    Soluo corrente melhorou", SQ->nS, SQ->WS); }
        return TRUE;
      }
    else if (cpk_lopt_same_set(nT, T, SQ->nS - nSred, &(SQ->S[nSred])))
      { /* Voltamos  soluo de partida. */
        if (verbose) { TRACE_SOL("    Ficou na mesma soluo", SQ->nS, SQ->WS); }
        return FALSE;
      }
    else
      { /* A emenda no foi melhor que o soneto, o ngcio  desfazer a troca. */
        int nDel = SQ->nS - nSred;
        if (verbose)
          {  if (cpk_weight_cmp(SQ->WS, WSold) < 0)
              { TRACE_SOL("      Soluo corrente piorou", SQ->nS, SQ->WS);
                int i;
                fprintf(stderr, "     ");
                for (i = 0; i < nDel; i++) 
                  { int w = SQ->S[nSred+i]; fprintf(stderr, " W[%d]=" WT_FFMT, w, SQ->W[w]); }
                fprintf(stderr, "\n");
                fprintf(stderr, "     ");
                for (i = 0; i < nT; i++) 
                  { int w = T[i]; fprintf(stderr, " W[%d]=" WT_FFMT, w, SQ->W[w]); }
                fprintf(stderr, "\n");
              }
            else 
              { TRACE_SOL("      Soluo mudou mas no melhorou", SQ->nS, SQ->WS); }
          }
        cpk_lopt_restore_solution(SQ, nDel, nT, T, verbose);
        if (verbose) { TRACE_SOL("    Soluo restaurada", SQ->nS, SQ->WS); }
        assert(cpk_weight_cmp(SQ->WS, WSold) == 0);
        return FALSE;
      }
  }
  
bool_t cpk_local_opt_at_pivot_2
  ( cpk_mis_state_t *SQ,     /* Soluo corrente {S} e estruturas auxiliares. */
    int nRem,        /* Nmero de vrtices a trocar em cada tentativa. */
    int x,           /* Vrtice piv. */
    bool_t verbose,  /* TRUE para imprimir mensagens de diagnstico. */
    bool_t piv[],    /* {piv[v] = TRUE} se vale a pena tentar {v} como piv. */
    int T[],
    bool_t temp[]
  )
  {
    cpk_graph_t *G = SQ->G;
    pqueue_t *Q = SQ->Q;
    
    assert(piv[x]); /* S deve chamar se {x} ainda est marcado como possvel piv. */
    assert(pqueue_count(Q) == 0); /* Neste ponto {S} deveria ser maximal. */

    /* O vrtice {x} realmente satisfaz as condies para piv? */
    if ((SQ->locS[x] >= 0) || (SQ->degS[x] != nRem)) { return FALSE; } 
    
    if (verbose) { fprintf(stderr, "    Mexendo nas redondezas do vrtice %d\n", x); }

    /* Lembra do peso da soluo corrente: */
    double WSold = SQ->WS;

    /* Retire de {S} todos os vizinhos de {x}, salvando-os em {T[0..nT-1]}. */
    /* Coloque os vrtices que ficaram admissveis em {Q}. */
    if (verbose) { fprintf(stderr, "      Retirando %d vrtices:", SQ->degS[x]); }
    int nT = 0;
    { int dx = G->deg[x]; int *nbx = &(G->nbr[G->fnb[x]]);
      int i;
      for (i = 0; i < dx; i++)
        { int v = nbx[i]; 
          if (SQ->locS[v] >= 0) 
            { cpk_mis_delete_indep(SQ, v, verbose); T[nT] = v; nT++; }
        }
    }
    assert(SQ->degS[x] == 0); /* Objetivo da retirada. */
    if (verbose) { fprintf(stderr, "\n"); }

    /* Lembra posio corrente em {S}: */
    int nSred = SQ->nS; 

    /* Coloca o vrtice {x} no conjunto {S} na marra: */
    if (verbose) { fprintf(stderr, "      Colocando %d na soluo...", x); }
    cpk_mis_add_indep(SQ, x, verbose);
    if (verbose) { fprintf(stderr, "\n");  }

    /* Roda algoritmo greedy sobre {G[Q]}: */
    if (verbose) { fprintf(stderr, "      Expanso gulosa..."); }
    cpk_greedy_maximize_lookahead(SQ, verbose, temp);
    if (verbose) { fprintf(stderr, " recolocou mais %d vrtices\n", SQ->nS - nSred);  }
    if (verbose) { TRACE_SOL("      Soluo corrente", SQ->nS, SQ->WS); }

    /* Marca os vrtices acrescentados como no mais pivotveis: */
    if (verbose) { fprintf(stderr, "      Eliminando futuros pivs:"); }
    { int i;
      for (i = nSred; i < SQ->nS; i++)
        { int s = SQ->S[i];
          piv[s] = FALSE; 
          if (verbose) { fprintf(stderr, " %d", s);}
        }
    }
    if (verbose) { fprintf(stderr, "\n");}

    /* Houve melhora? */
    if (cpk_weight_cmp(SQ->WS, WSold) > 0)
      { /* Houve melhora! */
        if (verbose) { TRACE_SOL("    Soluo corrente melhorou", SQ->nS, SQ->WS); }
        return TRUE;
      }
    else if (cpk_lopt_same_set(nT, T, SQ->nS - nSred, &(SQ->S[nSred])))
      { /* Voltamos  soluo de partida. */
        if (verbose) { TRACE_SOL("    Ficou na mesma soluo", SQ->nS, SQ->WS); }
        return FALSE;
      }
    else
      { /* A emenda no foi melhor que o soneto, o ngcio  desfazer a troca. */
        int nDel = SQ->nS - nSred;
        if (verbose)
          {  if (cpk_weight_cmp(SQ->WS, WSold) < 0)
              { TRACE_SOL("      Soluo corrente piorou", SQ->nS, SQ->WS);
                int i;
                fprintf(stderr, "     ");
                for (i = 0; i < nDel; i++) 
                  { int w = SQ->S[nSred+i]; fprintf(stderr, " W[%d]=" WT_FFMT, w, SQ->W[w]); }
                fprintf(stderr, "\n");
                fprintf(stderr, "     ");
                for (i = 0; i < nT; i++) 
                  { int w = T[i]; fprintf(stderr, " W[%d]=" WT_FFMT, w, SQ->W[w]); }
                fprintf(stderr, "\n");
              }
            else 
              { TRACE_SOL("      Soluo mudou mas no melhorou", SQ->nS, SQ->WS); }
          }
        cpk_lopt_restore_solution(SQ, nDel, nT, T, verbose);
        if (verbose) { TRACE_SOL("    Soluo restaurada", SQ->nS, SQ->WS); }
        assert(cpk_weight_cmp(SQ->WS, WSold) == 0);
        return FALSE;
      }
  }
  
    
