/* See cpk_lopt_simple.h */
/* Last edited on 2008-01-09 00:32:10 by stolfi */ 

#include <cpk_lopt_simple.h>
#include <cpk_lopt.h>
#include <cpk_greedy.h>
#include <cpk_graph.h>
#include <cpk_mis.h>
#include <cpk_weight.h>
#include <cpk_valid.h>
#include <cpk_debug.h>
#include <cpk_io.h>
#include <cpk_basic.h>

#include <bool.h>
#include <vec.h>

#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <assert.h>

/* PROTOTIPOS INTERNOS */

bool_t cpk_local_opt_twiddle_one
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int u,             /* Vrtice a retirar de {S}. */
    bool_t verbose,    /* TRUE para imprimir mensagens de diagnstico. */
    bool_t temp[]      /* rea de trabalho */
  );
  /* Tenta melhorar a soluo {S} eliminando o vrtice {u} de {S} e
    cobrindo o buraco de forma a deixar {S} maximal novamente. na
    sada, o peso total {WS} no deve ser menor do que na entrada.
    Devolve {TRUE} se e somente se {S} foi alterado.
    
    A rea {temp} deve ter pelo menos {nV} elementos; deve ser toda FALSE
    na entrada e ser toda FALSE na sada. */

/* IMPLEMENTAES */

void cpk_local_opt_simple
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int *nSM,          /* Tamanho da melhor soluo j vista. */
    int SM[],          /* {SM[0..nSM-1]}  a melhor soluo j vista. */
    double *WSM,       /* Peso total da melhor soluo {SM}. */
    bool_t verbose     /* TRUE para imprimir mensagens de diagnstico. */
  )
  {
    int nV = SQ->G->nV;
    /* Areas de trabalho: */
    bool_t temp[nV];
    { int v; for (v = 0; v < nV; v++) { temp[v] = FALSE; } }
    
    /* Repete o passo base at tentar todos os vrtices de {S} em seguida sem sucesso: */
    int iu = 0; /* ndice em {S} do ltimo vrtice experimentado: */
    int nTentou = 0; /* Nmero de tentativas fracassadas consecutivas. */
    int nPass = 0; /* Nmero de tentativas que deram certo. */
    do
      { if (nTentou == 0) { if (verbose) { fprintf(stderr, "\n  Passada %d\n\n", nPass); } }
        /* Ajusta {iu}: */
        iu--; if (iu < 0) { iu = SQ->nS-1; }
        /* Pega o prximo vrtice: */
        int u = SQ->S[iu];
        if (verbose) { fprintf(stderr, "  Tentando %d\n", nPass); }
        /* Lembra do peso atual: */
        double WSold = SQ->WS;
        int nSold = SQ->nS;
        /* Tente melhorar {S} retirando o vrtice {u}: */
        if (cpk_local_opt_twiddle_one(SQ, u, verbose, temp))
          { /* Mudou; deve ser para melhor: */
            assert(cpk_weight_cmp(WSold, SQ->WS) < 0);
            /* Tente atualizar a melhor soluo. */
            cpk_lopt_update_best_solution(SQ, nSM, SM, WSM, verbose);
            /* comece tudo de novo: */ 
            nTentou = 0; nPass++;
          }
        else
          { /* No mudou; deve ter mantido peso e tamanho: */
            assert(nSold == SQ->nS);
            assert(cpk_weight_cmp(WSold, SQ->WS) == 0);
            /* Conte mais um fracasso: */
            nTentou++;
          }
      }
    while (nTentou < SQ->nS);
  }

bool_t cpk_local_opt_twiddle_one
  ( cpk_mis_state_t *SQ, /* Soluo corrente {S} e estruturas auxiliares. */
    int u,             /* Vrtice a retirar de {S}. */
    bool_t verbose,    /* TRUE para imprimir mensagens de diagnstico. */
    bool_t temp[]      /* rea de trabalho */
  )
  { pqueue_t *Q = SQ->Q; 
    assert(SQ->locS[u] >= 0);
    /* Lembra tamanho e pesos originais: */
    int nSold = SQ->nS;
    double WSold = SQ->WS;
    /* Retira o vrtice {u} do conjunto {S}, atualizando {Q} e pesos: */
    if (verbose) { fprintf(stderr, "  Retirando vrtice %d:", u); }
    cpk_mis_delete_indep(SQ, u, verbose);
    if (verbose) { fprintf(stderr, "\n"); }
    assert(pqueue_count(Q) > 0);
    assert(SQ->nS == nSold-1);
    /* Expande {S} para maximal, gulosamente mas com lokahead de 2: */
    if (verbose) { fprintf(stderr, "  Expanso gulosa..."); }
    cpk_greedy_maximize_lookahead(SQ, verbose, temp);
    if (verbose) { fprintf(stderr, " recolocou %d vrtices\n", SQ->nS - (nSold-1)); }
    /* Consistncias: */
    assert(pqueue_count(Q) == 0);
    if (cpk_weight_cmp(SQ->WS, WSold) > 0)
      { if (verbose) { TRACE_SOL("  Soluo corrente melhorou", SQ->nS, SQ->WS); }
        return TRUE;
      }
    else if ((SQ->nS == nSold) && (SQ->S[SQ->nS-1] == u))
      { /* Pelas propriedades de {cpk_maximize_greedy_lookahead}, {S} no mudou: */
        if (verbose) { TRACE_SOL("  Voltou para mesma soluo", SQ->nS, SQ->WS); }
        return FALSE;
      }
    else 
      { /* Soluo mudou mas no melhorou. Restaure mesmo que igual, para evitar loops: */
        int v = SQ->S[SQ->nS-1]; /* Vrtice acrescentado. */
        if (verbose) 
          { if (cpk_weight_cmp(SQ->WS, WSold) < 0)
              { TRACE_SOL("  Soluo corrente piorou", SQ->nS, SQ->WS);
                fprintf(stderr, "   "); 
                fprintf(stderr, " W[%d] = " WT_FFMT, u, SQ->W[u]);
                fprintf(stderr, " W[%d] = " WT_FFMT, v, SQ->W[v]);
              }
            else 
              { TRACE_SOL("  Soluo mudou mas no melhorou", SQ->nS, SQ->WS); }
          }
        if (verbose) { fprintf(stderr, "    Retirando vrtice %d:", v); }
        cpk_mis_delete_indep(SQ, v, verbose);
        if (verbose) { fprintf(stderr, "\n"); }
        if (verbose) { fprintf(stderr, "    Recolocando vrtice %d:", u); }
        cpk_mis_add_indep(SQ, u, verbose);
        if (verbose) { fprintf(stderr, "\n"); }
        if (verbose) { TRACE_SOL("  Soluo restaurada", SQ->nS, SQ->WS); }
        assert(cpk_weight_cmp(SQ->WS, WSold) == 0);
        return FALSE;
      }   
  }
