bool_t dg_cell_intersects_star(dg_dim_t d, dg_cell_index_t k, dg_locus_t E)
  {
    dg_rank_t Er = dg_cell_rank(E.cell);
    dg_grid_pos_t Ep[dg_MAX_GRID_DIM];
    dg_cell_position(d, E.cell, Ep);
    
    /* If {k} is deeper than {E}, get its ancestor of same rank: */
    dg_rank_t kr = dg_cell_rank(k);
    if (kr > Er) { k = k >> (kr - Er); kr = Er; }
    /* Get posns of min and max descendants of {k} on level {Er}: */
    dg_grid_pos_t kp_lo[dg_MAX_GRID_DIM];
    dg_cell_position(d, k << (Er - kr), kp_lo);
    dg_grid_pos_t kp_hi[dg_MAX_GRID_DIM];
    dg_cell_position(d, ((k + 1) << (Er - kr)) - 1, kp_hi);

    /* Get grid sizes on level {Er}: */
    dg_grid_size_t sz[dg_MAX_GRID_DIM];
    dg_grid_size(d, Er, sz);
    
    int i;
    for (i = 0; i < d; i++)
      { dg_grid_pos_t Epi = Ep[i];
        if ((Epi < kp_lo[i]) || (Epi > kp_hi[i]))
          { /* Try adjacent cell of {K(E)}, if any: */
            if (box_axis_belongs(i, E.norm))
              { dg_grid_size_t m = sz[i];
                Epi = (Epi + m - 1) % m;
                if ((Epi < kp_lo[i]) || (Epi > kp_hi[i]))
                  { return FALSE; }
              }
            else
              { return FALSE; }
          }
      }
    return TRUE;
  }

dg_wing_index_t dg_cell_index_in_star(dg_dim_t d, dg_cell_index_t k, dg_locus_t E)
  {
    dg_rank_t k_rank = dg_cell_rank(k);
    dg_rank_t E_rank = dg_cell_rank(E.cell);
    if (k_rank != E_rank) { return dg_NOT_WING; }

    dg_grid_pos_t k_pos[dg_MAX_GRID_DIM];
    dg_cell_position(d, k, k_pos);
    
    dg_grid_pos_t E_pos[dg_MAX_GRID_DIM];
    dg_cell_position(d, E.cell, E_pos);
 
    int i, j;
    dg_wing_index_t ix = 0;
    for (i = 0, j = 0; i < d; i++)
      { if (box_axis_belongs(i, E.norm))
          { if (k_pos[i] != E_pos[i])
              { dg_grid_pos_t m = dg_max_grid_pos(d, E_rank, i);
                dg_grid_pos_t dp = (E_pos[i] + m - k_pos[i]) % m;
                if (dp != 1) { return dg_NOT_WING; }
                ix += (1 << j);
              }
            j++;
          }
        else
          { if (k_pos[i] != E_pos[i]) 
              { return dg_NOT_WING; }
          }
      }
    return ix;
  }

void dg_star_box_root_relative(dg_dim_t d, dg_locus_t E, interval_t B[])
  {
    int i;
    dg_cell_box_root_relative(d, E.cell, B);
    for (i = 0; i < d; i++)
      { if (box_axis_belongs(i, E.norm))
          { LO(B[i]) = 2*LO(B[i]) - HI(B[i]); }
      }
  }

dg_locus_vec_t dg_find_critical_loci(dg_dim_t d, dg_node_t *root, bool_t vtxOnly, bool_t maximal)
  { dg_locus_vec_t crit = dg_locus_vec_new(100);
    int nCrit = 0;

    auto void visit
      ( dg_locus_t E, 
        dg_rank_t r, 
        dg_node_star_t *NE, 
        dg_locus_data_t *dtE, 
        bool_t maxE
      );

    void visit
      ( dg_locus_t E, 
        dg_rank_t r, 
        dg_node_star_t *NE, 
        dg_locus_data_t *dtE, 
        bool_t maxE
      )
      { if ((maximal == maxE) && ((! vtxOnly) || (dg_locus_dimension(d, E) == 0)))
          { if (nCrit >= crit.nel) { dg_locus_vec_expand(&crit, nCrit); }
            crit.el[nCrit] = E;
            nCrit++;
          }
      }
      
    dg_enum_critical_loci(d, root, visit, NULL, NULL, NULL, NULL);
    dg_locus_vec_trim(&crit, nCrit);
    return crit;
  }

void dg_enum_critical_loci
  ( dg_dim_t d, 
    dg_node_t *root,
    dg_locus_visit_proc_t *visit, 
    dg_root_locus_data_proc_t *root_data,
    dg_shrink_locus_data_proc_t *shrink_data,
    dg_split_locus_data_proc_t *split_data,
    dg_free_locus_data_proc_t *free_data
  )
  { 
    auto void dg_enum_critical_sub_loci
      ( dg_locus_t E, 
        dg_rank_t r, 
        dg_node_star_t *NE, 
        dg_locus_data_t *dtE
      );

    void dg_enum_critical_sub_loci
      ( dg_locus_t E, 
        dg_rank_t r, 
        dg_node_star_t *NE, 
        dg_locus_data_t *dtE
      )
      {
        /* fprintf(stderr, "%*s [enter rank = %d locus = ", 2*r, "", r); */
        /* dg_locus_print(stderr, d, E); */
        affirm(dg_node_star_is_valid(r, NE), "invalid locus"); 
        if (dg_node_star_is_leaf(NE))
          { /* Leaf locus. */
            /* fprintf(stderr, " leaf"); */
            visit(E, r, NE, dtE, FALSE);
            /* fprintf(stderr, "]\n"); */
          }
        else
          { /* Locus is not a leaf locus; split cells of {K(E)} and recurse. */
            /* fprintf(stderr, " nonleaf: "); */
            dg_axis_t splax = dg_split_axis(d, r); /* Splitting axis. */
            /* Generate subtrees: */
            if (box_axis_belongs(splax, E.norm))
              { 
                /* The split axis is perpendicular to the item. */
                dg_axis_index_t splj = box_find_axis(E.norm, splax);
                affirm(splj < d, "inconsistent axis sets");
                /* fprintf(stderr, " shrink on axis %d = Nrm(E)[%d]\n", splax, splj); */                /* fprintf(stderr, "]\n"); */

                /* Get locus {E1} that is the same item in the next level: */
                dg_locus_t E1 = dg_locus(E.norm, 2*E.cell);  

                /* Shrink node star around element: */
                dg_node_ref_vec_t NE1;
                dg_node_star_shrink(NE, splj, &NE1);
                dg_locus_data_t *dtE1 = 
                  (shrink_data == NULL ? NULL : shrink_data(E, r, splax, NE, dtE));
                /* Recurse on same item in the next level, with same shape. */
                dg_enum_critical_sub_loci(E1, r+1, &NE1, dtE1);
                if (free_data != NULL) { free_data(E1, r+1, &NE1, dtE1); }
                free(NE1.el); 
              }
            else
              { /* The split axis is parallel to item {E}, split it: */
                /* fprintf(stderr, " bisect locus:"); */

                /* Compute normal axes of splitting item: */
                dg_axes_t NrmMD = box_axis_include(splax, E.norm); 

                /* Find rel position {splj} of {splax} among the {Nrm(EMD)}: */
                dg_axis_index_t splj = box_find_axis(NrmMD, splax);
                affirm(splj < d, "inconsistent axis sets");
                /* fprintf(stderr, " split on axis %d = Nrm(EMD)[%d]\n", splax, splj); */
                /* fprintf(stderr, "]\n"); */

                /* Pieces of this locus: */
                dg_locus_t ELO = dg_locus(E.norm, 2*E.cell);
                dg_locus_t EHI = dg_locus(E.norm, 2*E.cell + 1);
                dg_locus_t EMD = dg_locus(NrmMD, 2*E.cell + 1);

                /* Node stars of the pieces: */
                dg_node_ref_vec_t NLO, NHI, NMD;
                dg_locus_data_t *dtLO, *dtHI, *dtMD;

                dg_node_star_split(NE, splj, &NLO, &NHI, &NMD);
                if (split_data != NULL)
                  { split_data(E, r, splax, NE, dtE, &dtLO, &dtHI, &dtMD); }
                else
                  { dtLO = dtHI = dtMD = NULL; }

                /* Recurse for each half of item {F}: */
                visit(ELO, r+1, &NLO, dtLO, TRUE);
                dg_enum_critical_sub_loci(ELO, r+1, &NLO, dtLO);

                visit(EHI, r+1, &NHI, dtHI, TRUE);
                dg_enum_critical_sub_loci(EHI, r+1, &NHI, dtHI);

                /* Recurse for the separating item {S} of {F}. */
                visit(EMD, r+1, &NMD, dtMD, TRUE);
                dg_enum_critical_sub_loci(EMD, r+1, &NMD, dtMD);

                /* Reclaim temporary storage. */
                if (free_data != NULL) 
                  { free_data(EMD, r+1, &NMD, dtMD); 
                    free_data(ELO, r+1, &NLO, dtLO);  
                    free_data(EHI, r+1, &NHI, dtHI);
                  }
                free(NMD.el);
                free(NLO.el);
                free(NHI.el);
              }
          }
        /* fprintf(stderr, "%*s [exit rank = %d locus = ", 2*r, "", r); */
        /* dg_locus_print(stderr, d, E); */
        /* fprintf(stderr, "]\n"); */
      }

    affirm(root != NULL, "tree must be non-null"); 
    affirm(root->index = 1, "must start at root");
    /* Start recursion with root cell and its faces: */
    { /* Allocate a root {dg_node_star_t} of sufficient size: */
      int max_ncells = (1 << d);
      dg_wing_index_t ix;
      dg_node_ref_vec_t NE = dg_node_ref_vec_new(max_ncells);
      for (ix = 0; ix < max_ncells; ix++) { NE.el[ix] = root; }
      
      /* Enumerate the faces of the root cell: */
      dg_axes_t nrm;
      for (nrm = 0; nrm < (1<<d); nrm++)
        { dg_locus_t E = dg_locus(nrm,1);
          int m = box_axes_count(nrm); /* Co-dimension of {E}. */
          NE.nel = (1 << m);
          dg_locus_data_t *dtE = (root_data == NULL ? NULL : root_data(E, root));
          visit(E, 0, &NE, dtE, TRUE);
          dg_enum_critical_sub_loci(E, 0, &NE, dtE);
          if (free_data != NULL) { free_data(E, 0, &NE, dtE); }
        }
      free(NE.el);
    }
  }

bool_t dg_node_star_is_valid(dg_rank_t r, dg_node_star_t *N)
  { int i;
    for (i = 0; i < N->nel; i++)
      { dg_node_t *Ni = N->el[i];
        if ((Ni == NULL) || (dg_cell_rank(Ni->index) != r))
          { return FALSE; }
      }
    return TRUE;
  }

bool_t dg_node_star_is_leaf(dg_node_star_t *N)
  { int i;
    for (i = 0; i < N->nel; i++)
      { dg_node_t *Ni = N->el[i];
        if (Ni->nodes == 1) { return TRUE; }
      }
    return FALSE;
  }

