/* Enumeration of faces in a dyadic grid. */
/* Last edited on 2007-01-04 00:16:42 by stolfi */

/* The procedures in this module recursivly enumerate selected cell
  faces of a subtree of the infinite dyadic cell tree, keeping track
  of the shape of each face or cell (a tensor-style Bezier patch) and
  the corresponding node(s) in some finite cell tree. */

#ifndef dg_plotextra_H
#define dg_plotextra_H

#include <dg_grid.h>
#include <dg_tree.h>
#include <bz_patch.h>
#include <dg_plot.h>
#include <dgfaceshape.h>
#include <stdlib.h>

void dg_enum_flattened
  ( dg_FaceShape *sh,    /* Shape of face {F}. */
    double tol,          /* Flattening tolerance. */
    dg_rank_t rMax,        /* Never split below this rank. */
    dg_FaceProc process  /* Client face-processing procedure. */  
  );
  /* Recursively partitions some dyadic face {F} into one or more
    smaller dyadic faces, until these can all be approximated by a
    flat (degree 1) Bezier patches, without discontinuities, and with 
    error at most {tol}. Then calls {process} on each flattened part.

    Arguments {ax, r, ka, na} describe the face {F} and its adjacent
    cells, as in the specification of {dg_FaceProc} (q.v.). 
    
    Argument {sh} describes the geometry of face {F}, as in a
    {dg_FaceProc}, except that its degree {b->g} is arbitrary.
    This procedure may modify the contents of {sh}. */
 
#endif
