/* See mdg_pack.h */
/* Last edited on 2009-08-23 12:11:24 by stolfi */

#define _GNU_SOURCE
#include <affirm.h>
#include <bool.h>
#include <indexing.h>
#include <stdlib.h>
#include <assert.h>
#include <mdg_grid.h>
  
#include <mdg_pack.h>

/* CELL PACKS */

uint64_t mdg_pack_slot_count(mdg_dim_t d, mdg_grid_size_t psz[])
  { uint64_t n = 1;
    int i;
    for (i = 0; i < d; i++) { n *= psz[i]; }
    return n;
  }

void mdg_pack_cells
  ( mdg_dim_t d, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[], 
    mdg_cell_index_t k[]
  )
  { mdg_grid_pos_t rpos[d]; /* Relative position of cell in pack. */
    /* Indexing steps: */
    ix_step_t pst[d];
    pst[0] = 1;
    { mdg_axis_t i; for (i=1; i <d; i++) { pst[i] = pst[i-1]*psz[i-1]; } }
    ix_fill(d, rpos, 0);
    int nC = 0;
    do
      { k[nC] = mdg_cell_shift(d, b, rpos);  nC++; }
    while (! ix_next(d, rpos, psz, ix_order_F, pst, NULL, NULL, NULL, NULL, NULL));
  }

mdg_pack_slot_index_t mdg_pack_find_cell
  ( mdg_dim_t d, 
    mdg_cell_index_t k, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[]
  )
  {
    mdg_rank_t kr = mdg_cell_rank(k);
    mdg_rank_t br = mdg_cell_rank(b);
    if (kr != br) { return mdg_NOT_SLOT; }

    mdg_grid_pos_t kp[mdg_MAX_GRID_DIM];
    mdg_cell_position(d, k, kp);
    
    mdg_grid_pos_t bp[mdg_MAX_GRID_DIM];
    mdg_cell_position(d, b, bp);
 
    int i;
    int pkx = 1;
    mdg_pack_slot_index_t ix = 0;
    for (i = 0; i < d; i++)
      { demand(psz[i] > 0, "bad pack size");
        mdg_grid_pos_t gszi = mdg_max_grid_pos(d, br, i);
        /* Get relative position of {k} in pack along axis {i}: */
        mdg_grid_pos_t ei = (kp[i] - bp[i] + gszi) % gszi;
        if (ei >= psz[i]) { return mdg_NOT_SLOT; }
        ix += ei * pkx;
        pkx *= psz[i];
      }
    return pkx;
  }

bool_t mdg_pack_intersects_cell
  ( mdg_dim_t d, 
    mdg_cell_index_t k, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[]
  )
  {
    mdg_rank_t br = mdg_cell_rank(b);
    mdg_grid_pos_t bp[mdg_MAX_GRID_DIM];
    mdg_cell_position(d, b, bp);
    
    /* If {k} is deeper than {b}, get its ancestor of same rank: */
    mdg_rank_t kr = mdg_cell_rank(k);
    if (kr > br) { k = k >> (kr - br); kr = br; }
    /* Get posns of min and max descendants of {k} on level {Er}: */
    mdg_grid_pos_t kp_lo[mdg_MAX_GRID_DIM];
    mdg_cell_position(d, k << (br - kr), kp_lo);
    mdg_grid_pos_t kp_hi[mdg_MAX_GRID_DIM];
    mdg_cell_position(d, ((k + 1) << (br - kr)) - 1, kp_hi);

    /* Get grid sizes on level {br}: */
    mdg_grid_size_t bgsz[mdg_MAX_GRID_DIM];
    mdg_grid_size(d, br, bgsz);
    
    int i;
    for (i = 0; i < d; i++)
      { demand(psz[i] > 0, "bad pack size");
        mdg_grid_size_t bgszi = bgsz[i];
        /* Get rel position range {plop..phip} of pack in level {br}: */ 
        mdg_grid_pos_t plop = bp[i];
        mdg_grid_pos_t phip = (plop + psz[i] - 1) % bgszi;
        /* Make sure {phip >= plop}: */
        if (phip < plop) { phip += bgszi; }
        /* Get rel position range {klop..khip} of {k} in the same level: */
        mdg_grid_pos_t klop = kp_lo[i];
        mdg_grid_pos_t khip = kp_hi[i];
        assert(klop <= khip); /* A single cell can't wrap around. */
        /* Check whether the ranges intersect (with or without wrap-around): */
        if (klop < plop)
          { if ((khip < plop) && (phip < klop + bgszi)) { return FALSE; } }
        else if (plop < klop)
          { if (klop > phip) { return FALSE; }}
      }
    return TRUE;
  }

void mdg_pack_box_root_relative
  ( mdg_dim_t d, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[], 
    interval_t B[]
  )
  { mdg_cell_box_root_relative(d, b, B);
    int i;
    for (i = 0; i < d; i++)
      { demand(psz[i] > 0, "bad pack size");
        HI(B[i]) = HI(B[i]) + (psz[i]-1)*(HI(B[i]) - LO(B[i]));
      }
  }
