/* See udg_grid.h */
/* Last edited on 2009-08-23 12:11:56 by stolfi */

#define _GNU_SOURCE
#include <affirm.h>
#include <bool.h>
#include <indexing.h>
#include <stdlib.h>
#include <assert.h>
  
#include <udg_grid.h>

udg_grid_size_t udg_grid_size(udg_rank_t r)
  { return (1LL << r);  }

udg_grid_pos_t udg_max_grid_pos(udg_rank_t r)
  { return (1LL << r) - 1; }

sign_t udg_rank_cmp(udg_rank_t *x, udg_rank_t *y)
  { if ((*x) < (*y))
      { return -1; }
    else if ((*x) > (*y))
      { return +1; }
    else
      { return 0; }
  }
  
/* CELL INDICES */
  
udg_rank_t udg_cell_rank(udg_cell_index_t k)
  { int r = 0;
    while (k > (1LL << 16)) { k >>= 16; r+= 16; }
    while (k > (1LL << 4)) { k >>= 4; r+= 4; }
    while (k > 1) { k >>= 1; r+= 1; }
    return r;
  }

udg_grid_pos_t udg_cell_position(udg_cell_index_t k)
  { udg_cell_index_t t = ((1LL << udg_cell_rank(k)) >> 1);
    return k - t;
  }
  
udg_cell_index_t udg_cell_from_position(udg_rank_t r, udg_grid_pos_t pos)
  { /* First cell of layer {r}: */
    udg_cell_index_t k0 = (1LL << r);
    /* Number of cells {sz} in layer {r}: */
    udg_cell_index_t sz = (1LL << r);
    /* Displacement {t} from {k0} is {pos % sz}: */
    uint64_t t = pos & (sz - 1);
    /* Compute index: */
    return k0 + t;
  }

udg_cell_index_t udg_cell_shift(udg_cell_index_t k, udg_grid_pos_t dp)
  { udg_rank_t r = udg_cell_rank(k);
    /* First cell of layer {r}: */
    udg_cell_index_t k0 = (1LL << r);
    /* Number of cells {sz} in layer {r}: */
    udg_cell_index_t sz = (1LL << r);
    /* Displacement {t} from {k0} is {(k - k0 + dp) % sz}: */
    uint64_t t = (k - k0 + dp) & (sz - 1);
    /* Compute index: */
    return k0 + t;
  }

void udg_cell_interval_root_relative(udg_cell_index_t k, interval_t *B)
  { udg_grid_pos_t pos = udg_cell_position(k);
    udg_grid_size_t gsz = udg_grid_size(udg_cell_rank(k));
    double scale = 1.0/((double)gsz);
    LO((*B)) = scale*((double)pos);
    HI((*B)) = scale*((double)(pos + 1));
  }

/* PRINTOUT */

void udg_cell_print(FILE *wr, udg_cell_index_t k)
  { 
    udg_grid_pos_t kp = udg_cell_position(k);
    udg_rank_t kr = udg_cell_rank(k);
    fprintf(wr, "%d:", (int)kr);
    fprintf(wr, "%d", (int)kp);
  }
  
/* VECTORS OF CELL INDICES */

vec_typeimpl(udg_cell_index_vec_t,udg_cell_index_vec,udg_cell_index_t);
