/* See udg_pack.h */
/* Last edited on 2009-08-23 12:10:30 by stolfi */

#define _GNU_SOURCE
#include <affirm.h>
#include <bool.h>
#include <indexing.h>
#include <stdlib.h>
#include <assert.h>
#include <vec.h>

#include <udg_grid.h>
  
#include <udg_pack.h>

/* CELL PACKS */

void udg_pack_cells(udg_cell_index_t b, udg_grid_size_t psz, udg_cell_index_t k[])
  { udg_pack_slot_index_t ix = 0;
    for (ix = 0; ix < psz; ix++)
      { k[ix] = udg_cell_shift(b, ix); }
  }

udg_pack_slot_index_t udg_pack_find_cell
  ( udg_cell_index_t k, 
    udg_cell_index_t b,
    udg_grid_size_t psz
  )
  {
    /* Get ranks of {k} and {b}: */
    udg_rank_t kr = udg_cell_rank(k);
    udg_rank_t br = udg_cell_rank(b);
    if (kr != br) { return udg_NOT_SLOT; }

    /* Get position of {k} and {b} in that layer: */
    udg_grid_pos_t kp = udg_cell_position(k);
    udg_grid_pos_t bp = udg_cell_position(b);
 
    /* Get position of {k} relative to {b}, beware of wrap-around: */
    udg_grid_size_t gsz = udg_grid_size(br);
    udg_grid_pos_t e = (kp - bp + gsz) % gsz;
    if (e >= psz) { return udg_NOT_SLOT; }
    return (udg_pack_slot_index_t)e;
  }

bool_t udg_pack_intersects_cell
  ( udg_cell_index_t k, 
    udg_cell_index_t b,
    udg_grid_size_t psz
  )
  {
    /* Get rank of {b}, its position in that layer, and size of layer: */
    udg_rank_t br = udg_cell_rank(b);
    udg_grid_pos_t bp = udg_cell_position(b);
    udg_grid_size_t bgsz = udg_grid_size(br);
    
    /* Get rank of {k}: */
    udg_rank_t kr = udg_cell_rank(k);
    
    /* If {k} is deeper than {b}, get its ancestor of same rank: */
    if (kr > br) { k = k >> (kr - br); kr = br; }
    
    /* Get positions of min and max descendants of {k} on level {Er}: */
    udg_grid_pos_t kp_lo = udg_cell_position(k << (br - kr));
    udg_grid_pos_t kp_hi = udg_cell_position(((k + 1) << (br - kr)) - 1);
    assert(kp_lo <= kp_hi); /* A single cell can't wrap around. */

    /* Get rel position range {pp_lo..pp_hi} of pack in level {br}: */ 
    udg_grid_pos_t pp_lo = bp;
    udg_grid_pos_t pp_hi = (pp_lo + psz - 1) % bgsz;
    
    /* Make sure {pp_hi >= pp_lo}: */
    if (pp_hi < pp_lo) { pp_hi += bgsz; }
    
    /* Check whether the ranges intersect (with or without wrap-around): */
    if (kp_lo < pp_lo)
      { if ((kp_hi < pp_lo) && (pp_hi < kp_lo + bgsz)) { return FALSE; } }
    else if (pp_lo < kp_lo)
      { if (kp_lo > pp_hi) { return FALSE; }}
    
    return TRUE;
  }

void udg_pack_interval_root_relative
  ( udg_cell_index_t b,
    udg_grid_size_t psz, 
    interval_t *B
  )
  { udg_cell_interval_root_relative(b, B);
    HI((*B)) = HI((*B)) + (psz-1)*(HI((*B)) - LO((*B)));
  }

