#define PROG_NAME "test_interpolate"
#define PROG_DESC "test of {float_image_interpolate_C0.h} and {float_image_interpolate_C1.h}"
#define PROG_VERS "1.0"

/* Last edited on 2010-08-14 23:33:37 by stolfi */ 
/* Created on 2009-06-02 by J. Stolfi, UNICAMP */

#define test_interpolate_COPYRIGHT \
  "Copyright  2009  by the State University of Campinas (UNICAMP)"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include <float_image.h>
#include <float_image_test.h>
#include <float_pnm_image.h>
#include <float_pnm_image_io.h>
#include <float_image_interpolate.h>
#include <r2.h>
#include <bool.h>
#include <jsfile.h>
#include <affirm.h>
#include <indexing.h>

int main(int argn, char **argv);

void do_shift_tests(void);
void do_plot_tests(void);

void do_shift_test
  ( float_image_t *iimg, /* Input image */
    r2_t ish,            /* Amount already shifted. */
    r2_t dsh,            /* Amount to shift */
    int order,           /* Continuity order. */
    ix_reduction_t red,  /* Index reduction policy. */
    float_image_t *oimg  /* Output image. */
  );
  /* Shifts image {iimg} by {dsh} and puts it in {oimg}. Writes {osh}
   to disk. Assumes that {iimg} has already been shifted by {ish}. */
  
void do_plot_test
  ( float_image_t *img,  /* Test image. */
    int kx,              /* Column of pulse. */
    int ky,              /* Row of pulse. */
    ix_reduction_t red   /* Index reduction policy. */
  );
  /* Fills {img} with a blip at column {ix}, row {iy}; then writes a 
    {gnuplot/splot} file with the result of interpolating  that image 
    over its domain. */

float_image_t *get_test_image(int NC);

void write_image(r2_t *ish, r2_t *dsh, r2_t *osh, int order, ix_reduction_t red, float_image_t *img);
  /* Writes an image {img} that was shifted from {ish} by {dsh} to {osh}. 
    If {ish} is NULL assumes that it is the original image. */

int main (int argn, char **argv)
  {
    do_plot_tests();
    do_shift_tests();
    return 0;
  }
    
void do_plot_tests(void)
  {
    int NC = 1, NX = 6, NY = 7;
    int CX = NX/2, CY = NY/2;
    float_image_t *img = float_image_new(NC, NX, NY);
    ix_reduction_t red;
    for (red = ix_reduction_FIRST; red <= ix_reduction_LAST; red++)
      { do_plot_test(img, CX,   CY,   red);
      
        int k;
        for (k = 0; k < 2; k++)
          {
            do_plot_test(img, CX,     k,      red);
            do_plot_test(img, CX,     NY-1-k, red);
            do_plot_test(img, k,      CY,     red);
            do_plot_test(img, NX-1-k, CY,     red);

            do_plot_test(img, 0,      k,      red);
            do_plot_test(img, 0,      NY-1-k, red);
            do_plot_test(img, NX-1,   k,      red);
            do_plot_test(img, NX-1,   NY-1-k, red);
            
            if (k != 0)
              { do_plot_test(img, k,      0,      red);
                do_plot_test(img, NX-1-k, 0,      red);

                do_plot_test(img, k,      NY-1,   red);
                do_plot_test(img, NX-1-k, NY-1,   red);
              }
          }
      }
    float_image_free(img);
  }

void do_shift_tests(void)
  {
    /* Input image for testing: */
    int NC = 3;
    float_image_t *img0 = get_test_image(NC);
    int NX = img0->sz[1];
    int NY = img0->sz[2];

    write_image(NULL, NULL, NULL, 0, FALSE, img0);
    
    /* Output (shifted) images: */
    float_image_t *img1 = float_image_new(NC, NX, NY);
    float_image_t *img2 = float_image_new(NC, NX, NY);
    float_image_t *img3 = float_image_new(NC, NX, NY);
    float_image_t *img4 = float_image_new(NC, NX, NY);
    float_image_t *img5 = float_image_new(NC, NX, NY);

    int order;
    ix_reduction_t red;
    for (red = ix_reduction_FIRST; red <= ix_reduction_LAST; red++)
      { 
        for (order = -1; order <= 1; order++)
          { 
            r2_t tsh, dsh;

            /* Shift little by little: */
            tsh = (r2_t){{ 0.00, 0.00 }};
            dsh = (r2_t){{ 0.00, 0.00 }}; do_shift_test(img0, tsh, dsh, order, red, img1); r2_add(&dsh, &tsh, &tsh);
            dsh = (r2_t){{ 3.30, 2.25 }}; do_shift_test(img1, tsh, dsh, order, red, img2); r2_add(&dsh, &tsh, &tsh);
            dsh = (r2_t){{ 2.30, 3.50 }}; do_shift_test(img2, tsh, dsh, order, red, img3); r2_add(&dsh, &tsh, &tsh);
            dsh = (r2_t){{ 3.40, 3.25 }}; do_shift_test(img3, tsh, dsh, order, red, img4); r2_add(&dsh, &tsh, &tsh);

            /* Shift all at once: */
            tsh = (r2_t){{ 0.00, 0.00 }};
            dsh = (r2_t){{ 9.00, 9.00 }}; do_shift_test(img0, tsh, dsh, order, red, img5); r2_add(&dsh, &tsh, &tsh);
          }
      }
    float_image_free(img0);
    float_image_free(img1);
    float_image_free(img2);
    float_image_free(img3);
    float_image_free(img4);
    float_image_free(img5);
  }   

float_image_t *get_test_image(int NC)
  {
    char *fname = NULL;
    asprintf(&fname, "in/orig.%s", (NC == 3 ? "ppm" : "pgm"));
    bool_t isMask = FALSE; /* Assume pixels have a smooth distribution. */
    float_image_t *base = float_pnm_image_read(fname, isMask, 1/0.4500, 0.0327, TRUE, TRUE, FALSE);

    int BNX = base->sz[1];
    int BNY = base->sz[2];
    
    int NX = BNX - (BNX % 2);
    int NY = BNY - (1 - (BNX % 2));
    
    float_image_t *img = float_image_crop(base, 0,NC, 0,NX, 0,NY, 0.5000);
    int c;
    for (c = 0; c < NC; c++)
      { int ip;
        int CX = NX/2, CY = NY/2;
        for (ip = 0; ip < 10; ip++)
          { float_image_set_sample(img, c, 0,     ip,      1.00);
            float_image_set_sample(img, c, NX-1,  ip,      1.00);
            float_image_set_sample(img, c, 0,     NY-1-ip, 1.00);
            float_image_set_sample(img, c, NX-1,  NY-1-ip, 1.00);
            
            float_image_set_sample(img, c, CX,       NY-CY-ip, 1.00);
            float_image_set_sample(img, c, NX-CX,    CY+ip,    1.00);
            float_image_set_sample(img, c, CX-ip,    CY,       1.00);
            float_image_set_sample(img, c, NX-CX+ip, NY-CY,    1.00);
            
            float_image_set_sample(img, c, ip,      0,     1.00);
            float_image_set_sample(img, c, ip,      NY-1,  1.00);
            float_image_set_sample(img, c, NX-1-ip, 0,     1.00);
            float_image_set_sample(img, c, NX-1-ip, NY-1,  1.00);
          }
      }
    free(fname);
    float_image_free(base);
    return img;
  }

void do_plot_test
  ( float_image_t *img,  /* Test image. */
    int kx,              /* Column of pulse. */
    int ky,              /* Row of pulse. */
    ix_reduction_t red   /* inedx reduction policy. */
  )
  {
    fprintf(stderr, "plotting blip at (%3d+,%3d+)\n", kx, ky);
    
    /* Get image dimensions: */
    int NC = img->sz[0]; assert(NC == 1);
    int NX = img->sz[1];
    int NY = img->sz[2];
    
    /* Fill image with blip: */
    float_image_fill_channel(img, 0, 0.0);
    float_image_set_sample(img, 0, kx, ky, 1.0);
    
    /* Generate 2D plot: */
    char *fname = NULL;
    asprintf(&fname, "out/R%c-%04d-%04d.txt", "SERMP"[red], kx, ky);
    FILE *plot = open_write(fname, TRUE);
    int m = 8; /* Subsampling rate */
    int G = 3; /* Extra plot margin in pixels. */
    double eps = 1.0/1024; /* Fudge to detect discontinuities. */
    int ix, iy, dx, dy;
    for (iy = -G; iy <= NY+G; iy++)
      { for (dy = 0; dy <= m; dy++)
          { for (ix = -G; ix <= NX+G; ix++)
              { for (dx = 0; dx <= m; dx++)
                  { /* Get coordinates of interpolation point: */
                    double x = ix + (dx + eps)/(m + 2*eps);
                    double y = iy + (dy + eps)/(m + 2*eps);
                    /* Interpolate image: */
                    float vn = float_image_interpolate_sample(img, 0, x, y, -1, red);
                    float v0 = float_image_interpolate_sample(img, 0, x, y,  0, red);
                    float v1 = float_image_interpolate_sample(img, 0, x, y,  1, red);
                    /* Characteristic function of domain: */
                    int s = ((x >= 0) && (x <= NX) && (y >= 0) && (y <= NY));
                    /* Wrte to plot file: */
                    fprintf(plot, "%8.5f %8.5f %10.7f %10.7f %10.7f %d\n", x, y, vn, v0, v1, s);
                  }
              }
            /* Blank line between scanlines: */
            fprintf(plot, "\n");
          }
      }
    /* Cleanup: */
    fclose(plot);
    free(fname);
  }  

void do_shift_test
  ( float_image_t *iimg, /* Input image. */
    r2_t ish,            /* Amount already shifted. */
    r2_t dsh,            /* Amount to shift */
    int order,           /* Continuity order. */
    ix_reduction_t red,  /* Index reduction policy. */
    float_image_t *oimg  /* Output image. */
  )
  {
    r2_t osh;
    r2_add(&ish, &dsh, &osh);
    fprintf(stderr, "shifting image from ");
    r2_print(stderr, &ish);
    fprintf(stderr, " by ");
    r2_print(stderr, &dsh);
    fprintf(stderr, " to ");
    r2_print(stderr, &osh);
    
    /* Get image dimensions: */
    int NC = iimg->sz[0]; assert(oimg->sz[0] == NC);
    int NX = iimg->sz[1]; assert(oimg->sz[1] == NX);
    int NY = iimg->sz[2]; assert(oimg->sz[2] == NY);
    
    /* Allocate a pixel's worth of samples: */
    double v[NC];
    float f[NC];
    
    /* Scan pixels: */
    int ix, iy;
    for (iy = 0; iy < NY; iy++)
      { for (ix = 0; ix < NX; ix++)
          { /* Get coordinates of pixel center, relative to image center: */
            r2_t p = (r2_t){{ ix + 0.5, iy + 0.5 }};
            /* Get coordinates of shifted pixel: */
            r2_t q; r2_sub(&p, &dsh, &q);
            /* Interpolate image: */
            float_image_interpolate_pixel(iimg, q.c[0], q.c[1], order, red, v);
            /* Store into pixel: */
            int i; for (i = 0; i < NC; i++) { f[i] = v[i]; }
            float_image_set_pixel(oimg, ix, iy, f);
          }
      }
    /* Write the output image: */
    write_image(&ish, &dsh, &osh, order, red, oimg);
  }  

void write_image(r2_t *ish, r2_t *dsh, r2_t *osh, int order, ix_reduction_t red, float_image_t *img)
  {
    char *ext = (img->sz[0] == 3 ? "ppm" : "pgm");
    
    auto int shtoi(r2_t *sh, int j);
      /* Converts coordinate {j} of {sh} from {[0 _ 1]} to an integer in {0..1000}: */
    
    char *fname = NULL;
    if (ish == NULL)
      { asprintf(&fname, "out/orig.%s", ext); }
    else
      { asprintf
          ( &fname,
            "out/C%c-R%c--%04d-%04d--%04d-%04d--%04d-%04d.%s",
            "n01"[order+1],
            "SERMP"[red],
            shtoi(osh,0), shtoi(osh,1),
            shtoi(ish,0), shtoi(ish,1),
            shtoi(dsh,0), shtoi(dsh,1), ext
          );
      }
    bool_t isMask = FALSE; /* Assume pixels have a smooth distribution. */
    float_pnm_image_write(fname, img, isMask, 0.4500, 0.0327, TRUE, TRUE, TRUE);
    free(fname);
    return;
    
    int shtoi(r2_t *sh, int j)
      { return (int)floor(sh->c[j]*1000 + 0.5); }
  }
