/* Test of jspng.h, jspng_image.h */
/* Last edited on 2011-06-16 21:32:57 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>

#include <bool.h>
#include <jspnm.h>
#include <jspnm_image.h>

#include <jspng_image.h>

int main (int argc, char **argv);

void do_jspng_image_io_tests(char *prefix);
  /* Performs various image I/O tests, reading files
    with names "{prefix}-rd-{VAR}.png" and writing files
    with names "{prefix}-wr-{QQQ}-{VAR}.png", where 
    {VAR} is "gry" or "rgb" and {QQQ} is the 3-digit 
    PNG quality parameter. */

void do_jspng_image_io_test(char *prefix, int ik, int iq);
  /* Reads an image from file "{prefix}-rd-{VAR}.png" and writes the
    complemented image as files  with names "{prefix}-wr-{QQQ}-{VAR}.png".

    The {ik} paramter indicates the colorspace. Currently the only 
    valid choices are 0 (the space {JCS_GRAYSCALE} of {pnglib.h},
    in which case {VAR} is "GRAY") or 1 ({JCS_RGB}in which case {VAR}
    is "RGB").
    
    The {iq} parameter determines the PNG {quality} setting. When
    {iq} is 0 the quality is 100% (no compression). As {iq} grows the
    quality decreases (non-linearly). The {QQQ} field in utput
    filenames is the 3-digit decimal value of the PNG {quality}. */
    
void frobnicate_image(pnm_image_t *img, pnm_image_t *omg);  
  /* Copies {img} ino {omg}, complementing some pixels
    relative to {img.maxval}, flipping left-right and
    top-bottom. Also sets {omg.maxval = img.maxval}. */

int main (int argc, char **argv)
  { do_jspng_image_io_tests("out/test");
    return 0;
  }
  
void do_jspng_image_io_tests(char *prefix)
  { int iq, ik;
    for (iq = 0; iq < 5; iq ++)
      for (ik = 0; ik < 2; ik++)
        { do_jspng_image_io_test(prefix, ik, iq); }
  }
      
void do_jspng_image_io_test(char *prefix, int ik, int iq)
  { 
    /* Tag indicating format variant: */
    /* Filename extension: */
    char *kindtbl[2] = {"GRAY", "RGB"};
    char *xkind = kindtbl[ik]; 
    
    /* PNG quality parameter: */
    int quality = 100 - iq*iq;
    
    fprintf(stderr, "-----------------------------------------\n");
    fprintf(stderr, "testing prefix = %s kind = %s quality = %d\n", prefix, xkind, quality);
    fprintf(stderr, "\n");

    char *fname = NULL;
    asprintf(&fname, "%s-rd-%s.png", prefix, xkind);
    int kind;
    pnm_image_t *img = jspng_image_read(fname, TRUE, &kind);
    pnm_image_describe(stderr, fname, img);
    fprintf(stderr, "PNG colorspace = %d\n", kind);
    free(fname);

    pnm_image_t *omg = pnm_image_new(img->cols, img->rows, img->chns);
    frobnicate_image(img, omg);
    fprintf(stderr, "\n");

    fname = NULL;
    asprintf(&fname, "%s-wr-%03d-%s.png", prefix, quality, xkind);
    pnm_image_describe(stderr, fname, omg);
    jspng_image_write(fname, omg, quality, TRUE);
    free(fname);
    fprintf(stderr, "-----------------------------------------\n");
  }

void frobnicate_image(pnm_image_t *img, pnm_image_t *omg)
  {
    int cols = img->cols; assert(img->cols == omg->cols);
    int rows = img->rows; assert(img->rows == omg->rows);
    int chns = img->chns; assert(img->chns == omg->chns);

    omg->maxval = img->maxval;
    
    int skip = 20;

    int x, y, c;
    for (y = 0; y < rows; y++)
      { bool_t yrev = ((y >= skip) && (y < rows-skip));
        pnm_sample_t *ip = img->smp[y];
        pnm_sample_t *op = omg->smp[rows - 1 - y];
        int ik = 0;
        int ok = cols*chns - 1;
        for (x = 0; x < cols; x++)
          { bool_t xrev = ((x >= skip) && (x < cols-skip));
            bool_t rev = (xrev && yrev);
            for (c = 0; c < chns; c++)
              { op[ok] = ( rev ? img->maxval - ip[ik] : ip[ik]);
                ik++; ok--;
              }
          }
      }
  }
