#define PROG_NAME "test_transform"
#define PROG_DESC "test of {float_image_transform.h}"
#define PROG_VERS "1.0"

/* Last edited on 2010-08-14 23:29:45 by stolfi */ 
/* Created on 2007-07-11 by J. Stolfi, UNICAMP */

#define test_transform_COPYRIGHT \
  "Copyright  2007  by the State University of Campinas (UNICAMP)"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include <float_image.h>
#include <float_pnm_image.h>
#include <float_image_transform.h>
#include <float_image_test.h>
#include <r2.h>
#include <r2x2.h>
#include <bool.h>
#include <jsfile.h>
#include <affirm.h>

int main(int argn, char **argv);

void do_test
  ( float_image_t *iimg,
    float_image_test_generator_t *gen_proc,
    char *gen_name,
    float_image_t *oimg,
    fitr_transform_t *map_proc,
    char *map_name,
    bool_t avg,
    int order
  );

void write_image(char *gen_name, char *map_name, bool_t avg, int order, char *io, float_image_t *img);

int main (int argn, char **argv)
  {
    /* Choose a min wavelength {PMin} for test images (pixels, integer): */
    int PMin = 4;
    
    /* Input image for testing (size must be multiple of {3*Pmin}): */
    int ichns = 3, icols = 9*(3*PMin), irows = 9*(3*PMin);
    float_image_t *iimg = float_image_new(ichns, icols, irows);
    
    /* Output (transformed) image: */
    int ochns = ichns, ocols = 200, orows = 150;
    float_image_t *oimg = float_image_new(ochns, ocols, orows);

    /* Get input and output image dimensions, floated: */
    double iwx = (double)icols;
    double iwy = (double)irows;
    double owx = (double)ocols;
    double owy = (double)orows;

    /* Compute circumradii {iR,oR} of input and output images (in pixels): */
    double iR = 0.5*hypot(icols, irows);
    double oR = 0.5*hypot(ocols, orows);
    fprintf(stderr, "iR = %7.2f  oR = %7.2f\n", iR, oR);
    
    /* Compute inradii {iH,oH} of input and output images (in pixels): */
    double iH = 0.5*(icols < irows ? icols : irows);
    double oH = 0.5*(ocols < orows ? ocols : orows);
    fprintf(stderr, "iH = %7.2f  oH = %7.2f\n", iH, oH);
    
    /* Geometry transformations: */
    auto void map_ident(r2_t *p, r2_t *ip, r2x2_t *J, bool_t *debug);
    auto void map_shift(r2_t *p, r2_t *ip, r2x2_t *J, bool_t *debug);
    auto void map_magnf(r2_t *p, r2_t *ip, r2x2_t *J, bool_t *debug);
    auto void map_reduc(r2_t *p, r2_t *ip, r2x2_t *J, bool_t *debug);
    auto void map_quadr(r2_t *p, r2_t *ip, r2x2_t *J, bool_t *debug);
    auto void map_polar(r2_t *p, r2_t *ip, r2x2_t *J, bool_t *debug);

    int order;
    for (order = 0; order <= 1; order++)
      { /* Run some tests: */
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_ident, "1-ident", TRUE,  order);
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_magnf, "2-magnf", TRUE,  order);
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_shift, "3-shift", TRUE,  order);
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_reduc, "4-reduc", TRUE,  order);
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_quadr, "5-quadr", TRUE,  order);
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_quadr, "5-quadr", FALSE, order);
        do_test(iimg, &float_image_test_gen_stripes, "1-stripes", oimg, &map_polar, "6-polar", TRUE,  order);
        do_test(iimg, &float_image_test_gen_ripples, "2-ripples", oimg, &map_quadr, "5-quadr", TRUE,  order);
        do_test(iimg, &float_image_test_gen_checker, "3-checker", oimg, &map_quadr, "5-quadr", TRUE,  order);
      }
    
    return 0;
    
    /* DEFINITION OF THE GEOMETRIC MAPS */

    void map_ident(r2_t *op, r2_t *ip, r2x2_t *J, bool_t *debug)
      {
        /* The identity map (copies the image aligned at bottom left corner). */
        
        /* Set the input coords: */
        (*ip) = (*op);
        
        /* Define input points outside the domain as invalid: */
        bool_t invalid = FALSE;
        invalid |= ((ip->c[0] < 0) | (ip->c[0] > icols));
        invalid |= ((ip->c[1] < 0) | (ip->c[1] > irows));
        if (invalid) 
          { (*ip) = (r2_t){{ NAN, NAN }}; }
        else
          { if (J != NULL)
              { /* Set {J} to the the Jacobian of the {(ox,oy) --> (ix,iy)} map: */
                r2x2_ident(J);
              }
          }
        
        /* Debug some output pixel: */
        r2_t cp = (r2_t){{ floor(0.4*owx)+0.5, floor(0.4*owy)+0.5 }}; 
        (*debug) = (r2_dist(op, &cp) <= 0.5000001);
      }

    void map_reduc(r2_t *op, r2_t *ip, r2x2_t *J, bool_t *debug)
      {
        /* Reduces the input image by fixed factor along each axis. */
        
        /* Get the coordinates {(ox,oy)} of {op}, unscaled: */
        double ox = (double)op->c[0];
        double oy = (double)op->c[1];
        
        /* Define the output-to-input scale factors: */
        double sx = 3.0;
        double sy = 1.5;
        
        /* Magnify by inverse scale factors: */
        double ix = sx*ox;
        double iy = sy*oy;
        
        /* Set the input coords: */
        ip->c[0] = ix;
        ip->c[1] = iy;
        
        /* Define input points outside the domain as invalid: */
        bool_t invalid = FALSE;
        invalid |= ((ip->c[0] < 0) | (ip->c[0] > icols));
        invalid |= ((ip->c[1] < 0) | (ip->c[1] > irows));
        if (invalid)
          { (*ip) = (r2_t){{ NAN, NAN }}; }
        else
          { if (J != NULL)
              { /* Set {J} to the the Jacobian of the {(ox,oy) --> (ix,iy)} map: */
                J->c[0][0] = sx;  /* dix_dox */ 
                J->c[0][1] = 0.0; /* diy_dox */ 
                J->c[1][0] = 0.0; /* dix_doy */ 
                J->c[1][1] = sy;  /* diy_doy */ 
              }
          }
        
        /* Debug some output pixel: */
        r2_t cp = (r2_t){{ floor(0.4*owx)+0.5, floor(0.4*owy)+0.5 }}; 
        (*debug) = (r2_dist(op, &cp) <= 0.5000001);
      }
  
    void map_magnf(r2_t *op, r2_t *ip, r2x2_t *J, bool_t *debug)
      {
        /* Expands the input image by fixed factor along each axis. */
        
        /* Get the coordinates {(ox,oy)} of {op}, unscaled: */
        double ox = (double)op->c[0];
        double oy = (double)op->c[1];
        
        /* Define the output-to-input scale factors: */
        double sx = 5.0;
        double sy = 3.5;
        
        /* Magnify by the scale factors: */
        double ix = ox/sx;
        double iy = oy/sy;
        
        /* Set the input coords: */
        ip->c[0] = ix;
        ip->c[1] = iy;
        
        /* Define input points outside the domain as invalid: */
        bool_t invalid = FALSE;
        invalid |= ((ip->c[0] < 0) | (ip->c[0] > icols));
        invalid |= ((ip->c[1] < 0) | (ip->c[1] > irows));
        if (invalid)
          { (*ip) = (r2_t){{ NAN, NAN }}; }
        else
          { if (J != NULL)
              { /* Set {J} to the the Jacobian of the {(ox,oy) --> (ix,iy)} map: */
                J->c[0][0] = 1/sx;  /* dix_dox */ 
                J->c[0][1] = 0.0;   /* diy_dox */ 
                J->c[1][0] = 0.0;   /* dix_doy */ 
                J->c[1][1] = 1/sy;  /* diy_doy */ 
              }
          }
        
        /* Debug some output pixel: */
        r2_t cp = (r2_t){{ floor(0.4*owx)+0.5, floor(0.4*owy)+0.5 }}; 
        (*debug) = (r2_dist(op, &cp) <= 0.5000001);
      }
  
    void map_shift(r2_t *op, r2_t *ip, r2x2_t *J, bool_t *debug)
      {
        /* Shifts the input image by fractions of a pixel. */
        
        /* Get the coordinates {(ox,oy)} of {op}, unscaled: */
        double ox = (double)op->c[0];
        double oy = (double)op->c[1];
        
        /* Define the input-to-output image shifts: */
        double sx = 10 + 0.50000000;
        double sy = 10 + 0.33333333;
        
        /* Apply the shifts: */
        double ix = ox - sx;
        double iy = oy - sy;
        
        /* Set the input coords: */
        ip->c[0] = ix;
        ip->c[1] = iy;
        
        /* Define input points outside the domain as invalid: */
        bool_t invalid = FALSE;
        invalid |= ((ip->c[0] < 0) | (ip->c[0] > icols));
        invalid |= ((ip->c[1] < 0) | (ip->c[1] > irows));
        if (invalid)
          { (*ip) = (r2_t){{ NAN, NAN }}; }
        else
          { if (J != NULL)
              { /* Set {J} to the the Jacobian of the {(ox,oy) --> (ix,iy)} map: */
                J->c[0][0] = 1.0; /* dix_dox */ 
                J->c[0][1] = 0.0; /* diy_dox */ 
                J->c[1][0] = 0.0; /* dix_doy */ 
                J->c[1][1] = 1.0; /* diy_doy */ 
              }
          }
        
        /* Debug some output pixel: */
        r2_t cp = (r2_t){{ floor(0.4*owx)+0.5, floor(0.4*owy)+0.5 }}; 
        (*debug) = (r2_dist(op, &cp) <= 0.5000001);
      }
  
    void map_quadr(r2_t *op, r2_t *ip, r2x2_t *J, bool_t *debug)
      {
        /* Get the coordinates {(ox,oy)} of {op}, in {[0_1]} coordinates: */
        double ox = ((double)op->c[0])/owx;
        double oy = ((double)op->c[1])/owy;
        
        /* Compute the coordinates {(ix,iy)} of {ip}, in {[0_1]} coordinates: */
        double Ax_xx = +1.00;
        double Ax_xy = 00.00;
        double Ax_yy = 00.00;
        double Bx_x =  00.00;
        double Bx_y =  00.00;
        double Cx =    00.00;
        
        double Ay_xx = 00.00;
        double Ay_xy = 00.00;
        double Ay_yy = +1.00;
        double By_x =  00.00;
        double By_y =  00.00;
        double Cy =    00.00;
        
        double ix = Ax_xx*ox*ox + 2*Ax_xy*ox*oy + Ax_yy*oy*oy + Bx_x*ox + Bx_y*oy + Cx;
        double iy = Ay_xx*ox*ox + 2*Ay_xy*ox*oy + Ay_yy*oy*oy + By_x*ox + By_y*oy + Cy;

        /* Convert input coords to pixels, with overflow: */
        double mrg = 0.1;         /* Relative overflow along each edge. */
        double mag = 1.0 + 2*mrg; /* Scaling factor. */
        
        /* Set the input coords: */
        ip->c[0] = (mag*ix - mrg)*iwx ;
        ip->c[1] = (mag*iy - mrg)*iwy;
        
        /* Arbitrarily define some input pixels as invalid: */
        double bx = 0.6;
        double by = 0.6;
        bool_t invalid = (hypot((ix-bx)*iwx, (iy-by)*iwy) <= 6.0);
        
        /* Define also input points outside the domain as invalid: */
        invalid |= ((ip->c[0] < 0) | (ip->c[0] > icols));
        invalid |= ((ip->c[1] < 0) | (ip->c[1] > irows));
        if (invalid)
          { (*ip) = (r2_t){{ NAN, NAN }}; }
        else
          { if (J != NULL)
              { /* Set {J} to the the Jacobian of the {(ox,oy) --> (ix,iy)} map: */
                J->c[0][0] = 2*(Ax_xx*ox + Ax_xy*oy) + Bx_x; /* dix_dox */ 
                J->c[0][1] = 2*(Ay_xx*ox + Ay_xy*oy) + By_x; /* diy_dox */ 
                J->c[1][0] = 2*(Ax_xy*ox + Ax_yy*oy) + Bx_y; /* dix_doy */ 
                J->c[1][1] = 2*(Ay_xy*ox + Ay_yy*oy) + By_y; /* diy_doy */ 

                /* Adjust {J} to account for the input and output scalings: */
                J->c[0][0] *= mag*iwx/owx;  /* diX_doX */ 
                J->c[0][1] *= mag*iwy/owx;  /* diY_doX */ 
                J->c[1][0] *= mag*iwx/owy;  /* diX_doY */ 
                J->c[1][1] *= mag*iwy/owy;  /* diY_doY */ 
              }
          }

        /* Debug some output pixel: */
        r2_t cp = (r2_t){{ floor(0.8*owx)+0.5, floor(0.6*owy)+0.5 }}; 
        (*debug) = (r2_dist(op, &cp) <= 0.5000001);
      }
  
    void map_polar(r2_t *op, r2_t *ip, r2x2_t *J, bool_t *debug)
      {
        /* Get the coordinates {(ox,oy)} of {op}, in {[-1:+1]} coordinates: */
        double ox = ((double)op->c[0] - owx/2)/oH;
        double oy = ((double)op->c[1] - owy/2)/oH;
        
        /* Convert to polar coordinates: */
        double ot = atan2(oy, ox);
        double or2 = ox*ox + oy*oy;
        
        /* Compute the coordinates {(ix,iy)} of {ip}, in {[0_1]} coordinates: */
        double ix = 0.5*(ot/M_PI + 1);
        ix = ix - floor(ix);
        double iy = or2;

        /* Convert input coords to pixels: */
        ip->c[0] = ix*iwx;
        ip->c[1] = iy*iwy;
        
        /* Define input pixels outside the image as invalid: */
        bool_t invalid = FALSE;
        invalid |= ((ip->c[0] < 0) | (ip->c[0] > icols));
        invalid |= ((ip->c[1] < 0) | (ip->c[1] > irows));
        if (invalid)
          { (*ip) = (r2_t){{ NAN, NAN }}; }
        else
          { if (J != NULL)
              { /* Set {J} to the the Jacobian of the {(ox,oy) --> (ix,iy)} map: */
                J->c[0][0] = -oy/or2*(0.5/M_PI); /* dix_dox */ 
                J->c[0][1] = 2*ox;               /* diy_dox */ 
                J->c[1][0] = +ox/or2*(0.5/M_PI); /* dix_doy */ 
                J->c[1][1] = 2*oy;               /* diy_doy */ 

                /* Adjust {J} to account for the input and output scalings: */
                J->c[0][0] *= iwx/oH;  /* diX_doX */ 
                J->c[0][1] *= iwy/oH;  /* diY_doX */ 
                J->c[1][0] *= iwx/oH;  /* diX_doY */ 
                J->c[1][1] *= iwy/oH;  /* diY_doY */ 
              }
          }
        
        /* Debug some output pixel: */
        r2_t cp = (r2_t){{ floor(0.4*owx)+0.5, floor(0.4*owy)+0.5 }}; 
        (*debug) = (r2_dist(op, &cp) <= 0.5000001);
      }
      
  }

void do_test
  ( float_image_t *iimg,
    float_image_test_generator_t *gen_proc,
    char *gen_name,
    float_image_t *oimg,
    fitr_transform_t *map_proc,
    char *map_name,
    bool_t avg,
    int order
  )
  {
    fprintf(stderr, "creating input image \"%s\" ...\n", gen_name);
    float_image_test_paint(iimg, gen_proc);
    double def = 1/M_PI; /* An arbitrary value. */
    fprintf(stderr, "filling output image with %8.6f ...\n", def);
    float_image_fill(oimg, def);
    double undef = 1/M_SQRT2; /* Another arbitrary value. */
    fprintf(stderr, "applying map \"%s\", undef = %8.6f avg = %c...\n", map_name, undef, "FT"[avg]);
    fitr_transform(iimg, map_proc, undef, avg, order, oimg);
    /* Write the input and output images: */
    write_image(gen_name, map_name, avg, order, "in", iimg);
    write_image(gen_name, map_name, avg, order, "ot", oimg);
  }  

void write_image(char *gen_name, char *map_name, bool_t avg, int order, char *io, float_image_t *img)
  {
    char *fname = NULL;
    if (map_name == NULL) { map_name = "0-nomap"; }
    asprintf(&fname, "test-%s-%s-%s-%c-%d.ppm", map_name, gen_name, io, "FT"[avg], order);
    FILE *wr = open_write(fname, TRUE);
    int chns = img->sz[0];
    bool_t yup = TRUE, verbose = TRUE;
    bool_t isMask = FALSE; /* Assume uniform distr. of pixel values in encoding/decoding. */
    pnm_image_t *pimg = float_image_to_pnm_image(img, isMask, chns, NULL, NULL, NULL, 255, yup, verbose);
    pnm_image_fwrite(wr, pimg, FALSE);
    pnm_image_free(pimg);
    fclose(wr);
    free(fname);
  }
