/* haar.c */
/* Last edited on 2008-05-25 02:49:33 by stolfi */

#include <stdio.h>
#include <pnm_leo.h>
#include <wave_leo.h>

#define SQRT2 1.4142136

static Gray gray, recon;
static char *infn;
static char *topfn    = NULL;
static char *recfn    = NULL;
static float scale    = SQRT2;
static int   minscale = 0;
static int   quiet    = 0;
static int   ncoef    = 20;
static float cutoff   = -9999;

void UsageExit(void);
void ParseArgs(int argc, char **argv);
void status(char *msg);
int main(int argc, char **argv);

void UsageExit(void)
{
  printf("usage: \n" 
    "  haar <pgmfile> \\\n" 
    "    [-t<topfile>] \\\n" 
    "    [-n<#coef>] \\\n" 
    "    [-s<scale>] \\\n" 
    "    [-m] \\\n" 
    "    [-r<recfile> [-c<cutoff>]] \\\n" 
    "    [-quiet] \\\n" 
    "    [-freq] \\\n" 
    "    [-bxy]\n"
  );
  exit(1);
}


void ParseArgs(int argc, char **argv)
{
  char c, *s;

  if (argc < 2) UsageExit();

  infn = argv[1];

  argv += 2; argc -= 2;

  while (argc)
    {
      if (argv[0][0] != '-') UsageExit();

      c =  argv[0][1];
      s = &argv[0][2];
      switch (c)
	{
	case 'b': GraySetDecompMode(MODE_MIXED(s[0]-'0',s[1]-'0'));
	  break;
	case 't': topfn = s;
	  break;
	case 's': scale = atof(s);
	  break;
	case 'm': minscale = 1;
	  break;
	case 'r': recfn = s;
	  break;
	case 'c': cutoff = atof(s);
	  break;
	case 'n': ncoef = atoi(s);
	  if (ncoef < 0 || ncoef > MAXB) ncoef = 20;
	  break;
	case 'q': quiet = 1;
	  break;
	case 'f': GraySetDecompMode(MODE_FREQUENCY);
	  break;
	default:  UsageExit();
	}

      argc--; argv++;
    }
}


void status(char *msg)
{
  if (!quiet) printf("%s\n", msg);
}


int main(int argc, char **argv)
{
  static int bins[MAXB];
  int    i;
  Image *image, *imageout;
  FILE  *fp;

  ParseArgs(argc, argv);

  status("reading image");

  image = ReadPNM(infn);
  if (!image) UsageExit();

  GrayFromImage(&gray, image);

  status("decomposing");

  GrayDecomp(&gray);
  GrayCopy(&gray, &recon);

  status("computing stats");
  GrayDecompStats(&gray, scale, minscale, bins, ncoef);

  if (topfn)
    {
      fp = fopen(topfn, "w");
      for (i = 0; i < ncoef; i++)
	fprintf(fp, "%d\n", bins[i]);
      fclose(fp);
    }

  if (recfn)
    {
      status("reconstructing");
      imageout = NewImage(image->width, image->width);
      if (cutoff < 0)
	GrayZeroAllButNamed(&recon, bins, ncoef);
      else
	GrayZeroAllSmall(&recon, &gray, cutoff);
      GrayRecon(&recon);
      GrayToImage(&recon, imageout);
      WritePGM(recfn, imageout);
    }
  exit(0);
}
