/* match.c */
/* Last edited on 2008-05-25 02:35:25 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define MAXN  80		/* max number of chars in name */
#define MAXB  200		/* max number of coef per image */
#define MAXDB 2500		/* max number of database entries */

int   dbn = 0;
int   dbi[MAXDB][MAXB];
int   target[MAXB];
int   scores[MAXDB];
char  names[MAXDB][MAXN];
int   cutoff = 0;		/* don't report scores below this */
char *targetName;
char *dbName;
char *cacheName = NULL;
int   ncoef = 20;		/* number of coef from target */
int   Ncoef = 20;		/* number of coef from database */

void UsageExit(void);
void ParseArgs(int argc, char **argv);
void FailExit(char *msg);
void ReadStat(char *fn, int *data, int nc);
int InList(int m, int *list, int size);
int TotalMatches(int *x1, int s1, int *x2, int s2);
void ComputeScores(char *fn);
void WriteCache(char *fn);
FILE *ReadCache(char *fn);
void ReadDB(char *fn);
int main(int argc, char **argv);

void UsageExit(void)
{
  printf("usage: \n"
    "  match <target> <database> \\\n"
    "    [-n<#target-coef>] \\\n"
    "    [-N<#db-coef>] \\\n"
    "    [-c<cutoff>] \\\n"
    "    [-w<cache-db-file>\n"
    );
  exit(0);
}


void FailExit(char *msg)
{
  if (msg)
    fprintf(stderr, "fatal error: %s\n", msg);
  else
    fprintf(stderr, "fatal error!\n");

  exit(-1);
}


void ParseArgs(int argc, char **argv)
{
  char c, *s;

  if (argc < 3) UsageExit();

  targetName = argv[1];
  dbName     = argv[2];

  argv += 3; argc -= 3;

  while (argc)
    {
      if (argv[0][0] != '-') UsageExit();

      c =  argv[0][1];
      s = &argv[0][2];
      switch (c)
	{
	case 'c': cutoff = atoi(s);
	  break;
	case 'w': cacheName = s;
	  break;
	case 'n': ncoef = atoi(s);
	  if (ncoef < 0 || ncoef > MAXB) ncoef = 20;
	  break;
	case 'N': ncoef = atoi(s);
	  if (Ncoef < 0 || Ncoef > MAXB) Ncoef = 20;
	  break;
	default:  UsageExit();
	}

      argc--; argv++;
    }
}


void ReadStat(char *fn, int *data, int nc)
{
  int   i,d;
  FILE *sfile = fopen(fn, "r");

  if (!sfile) FailExit("couldn't open a database file");

  for (i = 0; i < nc; i++)
    {
      if (fscanf(sfile,"%d",&d) != 1) 
	FailExit("couldn't finish reading a database file");
      data[i] = d;
    }

  fclose(sfile);
}


int InList(int m, int *list, int size)
{
  int i;

  for (i = 0; i < size; i++)
    if (m == list[i]) return 1;

  return 0;
}


int TotalMatches(int *x1, int s1, int *x2, int s2)
{
  int i;
  int matches = 0;

  for (i = 0; i < s1; i++)
    matches += InList(x1[i], x2, s2);

  return matches;
}


void ComputeScores(char *fn)
{
  int i, score;

  for (i = 0; i < dbn; i++)
    {
      score = TotalMatches(target, ncoef, dbi[i], Ncoef);
      if (score >= cutoff)
	printf("%d\t%s\t%s\n",score,fn,names[i]);
    }
}


void WriteCache(char *fn)
{
  int dbiSize  = sizeof(dbi);
  int nameSize = sizeof(names);
  FILE *cache  = fopen(fn, "w");

  if (!cache) FailExit("couldn't open cache file");

  fprintf(cache, "$$ cache for database used in `match'\n");

  fprintf(cache, "%d %d %d %d\n", dbn, Ncoef, dbiSize, nameSize);

  fwrite((char *) &dbi[0][0],   dbiSize,  1, cache);
  fwrite((char *) &names[0][0], nameSize, 1, cache);

  fclose(cache);
}


/*
 * returns NULL if successful, file pointer if this is not a cache.
 */
FILE *ReadCache(char *fn)
{
  static char buf[120];
  int nc, dbis, ns;
  int dbiSize  = sizeof(dbi);
  int nameSize = sizeof(names);
  FILE *cache  = fopen(fn, "r");

  if (!cache) FailExit("couldn't open database file");

  fgets(buf, 120, cache);

  if (buf[0] != '$' || buf[1] != '$') /* not a cache */
    {
      rewind(cache);
      return cache;
    }

  fgets(buf, 120, cache);
  sscanf(buf, "%d%d%d%d", &dbn, &nc, &dbis, &ns);

  if (dbn > MAXDB)     FailExit("cache size too big");
  if (Ncoef > nc)      FailExit("not that many coef in cache");
  if (dbis != dbiSize) FailExit("cache coef size differs from allocated mem");
  if (ns != nameSize)  FailExit("cache name size differs from allocated mem");

  fread((char *) &dbi[0][0],   dbiSize,  1, cache);
  fread((char *) &names[0][0], nameSize, 1, cache);

  fclose(cache);

  return NULL;
}


void ReadDB(char *fn)
{
  static char buf[MAXN];
  FILE *db = ReadCache(fn);

  if (!db) return;		/* read from cache already */

  while(!feof(db))
    {
      if (fscanf(db, "%s", buf) == 1)
	{
	  strcpy(names[dbn],buf); /* keep the name */

	  ReadStat(buf, dbi[dbn], Ncoef);

	  if (dbn < MAXDB) 
	    dbn++;
	  else
	    FailExit("not enough room for database\n");
	}
    }
  fclose(db);
}


int main(int argc, char **argv)
{
  ParseArgs(argc, argv);

  ReadDB(dbName);

  printf("\n___%s___\n", targetName);
  ReadStat(targetName, target, ncoef);

  if (cacheName) 
    WriteCache(cacheName);

  ComputeScores(targetName);
  return 0;
}

