#ifndef psp_grid_H
#define psp_grid_H

/* Regular grids, infinite or toroidal. */
/* Last edited on 2011-09-19 01:03:30 by stolfilocal */

#define _GNU_SOURCE
#include <stdint.h>
#include <stdio.h>

#include <interval.h>
#include <box.h>
#include <bz_basic.h>
#include <bool.h>
#include <sign.h>
#include <vec.h>

typedef rgrid_size_t psp_grid_size_t;
  /* Size of a circular grid, in multiples of the grid step.
    By convention, the value 0 denotes an infinite size. */

typedef rgrid_pos_t psp_grid_pos_t;
  /* Displacement of an item in a regular unidimensional grid, or a relative
    displacement between two items; both in multiples of the grid step. Also. */

typedef rgrid_dim_t psp_dim_t;
  /* Dimension of a domain or range space, mesh element, etc.. */
  
typedef rgrid_axis_t psp_axis_t; 
  /* Identifies a coordinate axis of a multidimensional 
    space, from 0 to {d-1} where {d} is the space's dimension. */

typedef rgrid_axis_set_t psp_axis_set_t; 
  /* A packed subset of coordinate axes (each a {psp_axis_t}). */

typedef rgrid_axis_index_t psp_axis_index_t; 
  /* Identifies a coordinate axis among a set of axes: 0 is the lowest. */

#define psp_NO_AXIS (rgrid_NO_AXIS)
  /* A {psp_axis_t} value that means "no axis". */

#define psp_AXIS_NOT_FOUND (RGRID_AXIS_NOT_FOUND)
  /* A {psp_axis_t} value that means "no such axis" */
  
psp_axis_set_t psp_axis_set_complement(psp_dim_t d, psp_axis_set_t A);
  /* The complement of {A} relative to the set {0..d-1}. */


#endif
