/* See pst_weight_map.h */
/* Last edited on 2010-05-04 01:39:05 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include <float_image.h>
#include <float_image_mscale.h>
#include <wt_table.h>

#include <pst_basic.h>
#include <pst_weight_map.h>
 
/* IMPLEMENTATIONS */

float_image_t *pst_weight_map_shrink(float_image_t *IW, bool_t harmonic, int avgWidth)
  { 
    demand(IW->sz[0] == 1, "weight map is not mono");
    int NX_JW = (IW->sz[1]+1)/2;
    int NY_JW = (IW->sz[2]+1)/2;
    int dxy = (avgWidth-1)/2;
    return float_image_mscale_mask_shrink(IW, NX_JW, NY_JW, dxy, dxy, avgWidth, harmonic);
  }

float_image_t *pst_weight_map_slope_to_height(float_image_t *W, bool_t harmonic, int NXV, int NYV)
  { 
    demand(W->sz[0] == 1, "weight map is not mono");
    int NXW = W->sz[1];
    int NYW = W->sz[2];
    
    /* Decide the window width {nw}, either 2 or 3: */
    int nw;
    if ((NXV == NXW) && (NYV == NYW))
      { /* Same size, average {3x3} blocks: */
        nw = 3;
      }
    else if ((NXV == NXW + 1) && (NYV == NYW + 1))
      { /* New maks is one larger, average {2x2} blocks: */
        nw = 2;
      }
    else
      { demand(FALSE, "wrong weight map size"); }

    /* Get the weight table: */
    double wt[nw];
    wt_table_fill_binomial(nw, wt);
    
    /* Compute displacement of window: */
    int dxy = (nw - 1)/2;
    
    /* Alocate the output image: */
    float_image_t *V = float_image_new(1, NXV, NYV);
    
    int x, y;
    for (y = 0; y < NYV; y++)
      { for (x = 0; x < NXV; x++)
          { double sum_cw = 0.0; /* Sum of window weights times {W} samples. */
            double sum_c = 0.0;  /* Sum of window weights. */
            int r, s;
            for (r = 0; r < nw; r++)
              { for (s = 0; s < nw; s++) 
                  { /* Get the sample from {W}: */
                    int xx = x + r - dxy;
                    int yy = y + s - dxy;
                    float wrs = 
                      ( (xx >= 0) && (xx < NXW) && (yy >= 0) && (yy < NYW) ?
                        float_image_get_sample(W, 0, xx, yy) :
                        0.0
                      );
                    demand(wrs >= 0.0, "negative weight");
                    demand(! isnan(wrs), "weight is NAN");
                    if (harmonic) { wrs = 1.0/wrs; }
                    /* Get the window weight {c}: */
                    double c = wt[r]*wt[s];  /* Window weight. */
                    /* Accumulate: */
                    sum_cw += c*wrs;
                    sum_c += c;
                  }
              }
            double avg = (harmonic ? sum_c/sum_cw : sum_cw/sum_c);
            float_image_set_sample(V, 0, x, y, avg);
          }
      }
    return V;
  }
