/* Last edited on 2008-05-09 10:23:50 by stolfi */
/* Quick Heap Sort - 
   
   see:
   Domenico Cantone, Gianluca Cincotti: 
   QuickHeapsort, an Efficient Mix of Classical Sorting Algorithms
   
   in:
   G. Bongiovanni, G. Gambosi, R. Petreschi (Eds.):
   CIAC 2000, LNCS 1767, 
   pp. 150-162,
   2000
   (c) Springer-Verlag Berlin Heidelberg 2000
*/


#include <intsort_extra2.h>

#define HEAP_TYPE_MIN 0
#define HEAP_TYPE_MAX 1

int path[LOG_ELEMENTS];

// we use the bottom up heapsort functions to build the heap used by the external heap sort...

inline int isrt_es_clever_quick_heap_LeafSearch(int root, int heapstart, int treeSize, int min_or_max) {
  int j, i=0;
  path[i++]=j=root;

  if( min_or_max == HEAP_TYPE_MIN ) 
    while(2*(j-heapstart+1) < treeSize)  {
      if (Compare_less(a[heapstart + 2*(j-heapstart+1)], a[heapstart + 2*(j-heapstart+1) - 1]))  // rson < lson?
	path[i++] = j = heapstart + 2*(j-heapstart+1);
      else
	path[i++] = j = heapstart + 2*(j-heapstart+1) - 1;
    }

  else
    while(2*(j-heapstart+1) < treeSize)  {
      if (Compare_less(a[heapstart + 2*(j-heapstart+1)], a[heapstart + 2*(j-heapstart+1) - 1])) // rson < lson?
	path[i++] = j = heapstart + 2*(j-heapstart+1) - 1;
      else
	path[i++] = j = heapstart + 2*(j-heapstart+1);
    }

  if (2*(j-heapstart+1) == treeSize)
    path[i++]=j=heapstart + treeSize - 1;

  return j;
}

inline int isrt_es_clever_quick_heap_BottomUpSearch(int root, int leaf, int heapstart, int min_or_max) {

  if( min_or_max == HEAP_TYPE_MIN ) 
    while((leaf>root) && Compare_less(a[root],a[leaf])) 
      leaf = heapstart + (leaf-heapstart+1)/2 - 1;

  else 
    while((leaf>root) && Compare_less(a[leaf],a[root]))
      leaf = heapstart + (leaf-heapstart+1)/2 - 1;
  
  return leaf;
}

inline void isrt_es_clever_quick_heap_Interchange(int i,int j) {
  int k,v;

  v = Assign(a[path[0]]);
  for(k=0;path[k]<j;k++) { 
    a[path[k]] = Assign(a[path[k+1]]);
  }
  a[path[k]] = Assign(v);
}


inline void isrt_es_clever_quick_heap_ReHeap(int root, int heapstart, int treeSize, int min_or_max) {
   int j;

   j = LeafSearch(root, heapstart, treeSize, min_or_max);
   j = BottomUpSearch(root, j, heapstart, min_or_max);
   Interchange(root,j);
}

// build a heap with a[left]...a[right], using the bottom-up strategy
inline void isrt_es_clever_quick_heap_build_heap( int left, int right, int min_or_max ) {
  int i;

  for (i = left+(right-left+1)/2;i>=left;i--) {
    ReHeap(i, left, (right-left+1), min_or_max);
  }
}


inline int isrt_es_clever_quick_heap_special_leaf( int left, int right, int min_or_max ) {
  int i;

  i = left+1;  // left son of root

  if( min_or_max == HEAP_TYPE_MAX )
    while( i<right ) {
      
      if( Compare_less(a[i], a[i+1]) ) 
	i++;
      
      //copy son to parent
      a[left + (i-left+1)/2 - 1] = a[i];
      
      // go to left son
      i = left + 2*(i-left+1) - 1;
    }
  else
    while( i<right ) {
      
      if( Compare_less(a[i+1], a[i]) ) 
	i++;
      
      //copy son to parent
      a[left+(i-left+1)/2 - 1] = a[i];
      
      // go to left son
      i = left + 2*(i-left+1) - 1;
    }

  
  if( i == right ) {
    a[left+(i-left+1)/2 - 1] = a[i];
    i = left + 2*(i-left+1) - 1;
  }

  return left + (i-left)/2;
}

void isrt_es_clever_quick_heap_external_minheap_sort(int heapleft, int heapright, int workleft) {
  int j, l, tmp;

  build_heap( heapleft, heapright, HEAP_TYPE_MIN );
  for( j=heapright; j>=heapleft; j-- ) {

    tmp=a[workleft];
    a[workleft] = a[heapleft]; 
    workleft++;

    l = special_leaf( heapleft, heapright, HEAP_TYPE_MIN );
    a[l] = tmp;
  }
}

void isrt_es_clever_quick_heap_external_maxheap_sort(int heapleft, int heapright, int workright) {
  int j, l, tmp;

  build_heap( heapleft, heapright, HEAP_TYPE_MAX );
  for( j=heapright; j>=heapleft; j-- ) {

    tmp=a[workright];
    a[workright] = a[heapleft]; 
    workright--;

    l = special_leaf( heapleft, heapright, HEAP_TYPE_MAX );
    a[l] = tmp;
  }
}


void isrt_es_clever_quick_heap_threesort(int left, int right) { 
  int i,j;
  double v;
  if (right-left == 1) {
    if (Compare_less(a[right],a[left])) Swap(left,right);
  } 
  else { 
    if (right-left == 2) {
      if (Compare_less(a[left],a[left+1])) 
        if (Compare_less(a[left+1],a[right])) 
          ;                       /* left < left+1 < right */ 
        else 
          if (Compare_less(a[left],a[right]))
            Swap(left+1,right);      /* left < right < left+1 */
          else  {
            v = Assign(a[left]);    /* right < left < left+1 */
            a[left] = Assign(a[right]);
            a[right] = Assign(a[left+1]);
            a[left+1] = Assign(v);
          }
      else
        if (Compare_less(a[right],a[left+1])) 
          Swap(left,right);      /* right < left+1 < left */ 
        else 
          if (Compare_less(a[left],a[right]))
            Swap(left+1,left);      /* left+1 < left < right */
          else  {
            v = Assign(a[left]);    /* left+1 < right < left */
            a[left] = Assign(a[left+1]);
            a[left+1] = Assign(a[right]);
            a[right] = Assign(v);
          }
    }
  }
}

void isrt_es_clever_quick_heap_cleverquickheap(int left, int right) { 
  int i,j,v;
  
  if (right-left >= 3) {
     if (Compare_less(a[left+1],a[right])) Swap(left+1,right);
     if (Compare_less(a[left],a[right])) Swap(left,right);
     if (Compare_less(a[left+1],a[left])) Swap(left+1,left);     
     i = left+1; j = right; v = Assign(a[left]);
     do { 
       do j--; while (Compare_less(a[j],v)); 
       do i++; while (Compare_less(v,a[i]));
       if (j >= i) Swap(i,j);
     } 
     while(j >= i);
     Swap(left, right - (j-left));
     if( (j-left) >= (right-j) ){
       external_minheap_sort( j+1, right, left );
       left = right - (j-left) + 1;
     }
     else {
       external_maxheap_sort( left+1, j, right );
       right = right - (j-left + 1);
     }
     cleverquickheap(left,right);
  }
  else  /* insertionsort(left,right); */ 
    threesort(left,right);
}

void isrt_es_clever_quick_heap_Sort() {

  cleverquickheap(0,NUM_ELEMENTS-1);
}
